#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for GOES 8-15 imager data in netCDF format.

Supports netCDF files from both NOAA-CLASS and EUMETSAT.



NOAA-CLASS
==========

GOES-Imager netCDF files from NOAA-CLASS contain detector counts alongside
latitude and longitude coordinates.

.. note ::
   If ordering files via NOAA CLASS, select 16 bits/pixel.


.. note ::
   Some essential information are missing in the netCDF files:

      1. Subsatellite point
      2. Calibration coefficients
      3. Detector-scanline assignment, i.e. information about which scanline
         was recorded by which detector

   Items 1. and 2. are not critical because the images are geo-located and NOAA
   provides static calibration coefficients (`[VIS]`_, `[IR]`_). The
   detector-scanline assignment however cannot be reconstructed properly. This
   is where an approximation has to be applied (see below).



Oversampling
------------

GOES-Imager oversamples the viewed scene in E-W direction by a factor of
1.75: IR/VIS pixels are 112/28 urad on a side, but the instrument samples
every 64/16 urad in E-W direction (see `[BOOK-I]`_ and `[BOOK-N]`_). That means
pixels are actually overlapping on the ground. This cannot be represented
by a pyresample area definition.

For full disk images it is possible to estimate an area definition with uniform
sampling where pixels don't overlap. This can be used for resampling and is
available via ``scene[dataset].attrs["area_def_uni"]``. The pixel size is derived
from altitude and N-S sampling angle. The area extent is based on the maximum
scanning angles at the earth's limb.


Calibration
-----------

Calibration is performed according to `[VIS]`_ and `[IR]`_, but with an average
calibration coefficient applied to all detectors in a certain channel. The
reason for and impact of this approximation is described below.

The GOES imager simultaneously records multiple scanlines per sweep using
multiple detectors per channel. The VIS channel has 8 detectors, the IR
channels have 1-2 detectors (see e.g. Figures 3-5a/b, 3-6a/b and 3-7/a-b in
`[BOOK-N]`_). Each detector has its own calibration coefficients, so in order to
perform an accurate calibration, the detector-scanline assignment is needed.

In theory it is known which scanline was recorded by which detector
(VIS: 5,6,7,8,1,2,3,4; IR: 1,2). However, the plate on which the detectors are
mounted flexes due to thermal gradients in the instrument which leads to a N-S
shift of +/- 8 visible or +/- 2 IR pixels. This shift is compensated in the
GVAR scan formation process, but in a way which is hard to reconstruct
properly afterwards. See `[GVAR]`_, section 3.2.1. for details.

Since the calibration coefficients of the detectors in a certain channel only
differ slightly, a workaround is to calibrate each scanline with the average
calibration coefficients. A worst case estimate of the introduced error can
be obtained by calibrating all possible counts with both the minimum and the
maximum calibration coefficients and computing the difference. The maximum
differences are:

======= ===== ====
GOES-8
------------------
Channel Diff  Unit
======= ===== ====
00_7    0.0   %  # Counts are normalized
03_9    0.187 K
06_8    0.0   K  # only one detector
10_7    0.106 K
12_0    0.036 K
======= ===== ====

======= ===== ====
GOES-9
------------------
Channel Diff  Unit
======= ===== ====
00_7    0.0   %  # Counts are normalized
03_9    0.0   K  # coefs identical
06_8    0.0   K  # only one detector
10_7    0.021 K
12_0    0.006 K
======= ===== ====

======= ===== ====
GOES-10
------------------
Channel Diff  Unit
======= ===== ====
00_7    1.05  %
03_9    0.0   K  # coefs identical
06_8    0.0   K  # only one detector
10_7    0.013 K
12_0    0.004 K
======= ===== ====

======= ===== ====
GOES-11
------------------
Channel Diff  Unit
======= ===== ====
00_7    1.25  %
03_9    0.0   K  # coefs identical
06_8    0.0   K  # only one detector
10_7    0.0   K  # coefs identical
12_0    0.065 K
======= ===== ====

======= ===== ====
GOES-12
------------------
Channel Diff  Unit
======= ===== ====
00_7    0.8   %
03_9    0.0   K  # coefs identical
06_5    0.044 K
10_7    0.0   K  # coefs identical
13_3    0.0   K  # only one detector
======= ===== ====

======= ===== ====
GOES-13
------------------
Channel Diff  Unit
======= ===== ====
00_7    1.31  %
03_9    0.0   K  # coefs identical
06_5    0.085 K
10_7    0.008 K
13_3    0.0   K  # only one detector
======= ===== ====

======= ===== ====
GOES-14
------------------
Channel Diff  Unit
======= ===== ====
00_7    0.66  %
03_9    0.0   K  # coefs identical
06_5    0.043 K
10_7    0.006 K
13_3    0.003 K
======= ===== ====

======= ===== ====
GOES-15
------------------
Channel Diff  Unit
======= ===== ====
00_7    0.86  %
03_9    0.0   K  # coefs identical
06_5    0.02  K
10_7    0.009 K
13_3    0.008 K
======= ===== ====


EUMETSAT
========

During tandem operations of GOES-15 and GOES-17, EUMETSAT distributed a
variant of this dataset with the following differences:

1. The geolocation is in a separate file, used for all bands
2. VIS data is calibrated to Albedo (or reflectance)
3. IR data is calibrated to radiance.
4. VIS data is downsampled to IR resolution (4km)
5. File name differs also slightly
6. Data is received via EumetCast


References:
===========

- `[GVAR]`_  GVAR transmission format
- `[BOOK-N]`_ GOES-N databook
- `[BOOK-I]`_ GOES-I databook (broken)
- `[IR]`_ Conversion of GVAR Infrared Data to Scene Radiance or Temperature
- `[VIS]`_ Calibration of the Visible Channels of the GOES Imagers and Sounders
- `[GLOSSARY]`_ GVAR_IMG Glossary
- `[SCHED-W]`_ GOES-15 Routine Imager Schedule
- `[SCHED-E]`_ Optimized GOES-East Routine Imager Schedule


.. _[GVAR]: https://noaasis.noaa.gov/NOAASIS/pubs/nesdis82.PDF
.. _[BOOK-N]: https://www.nasa.gov/pdf/148080main_GOES-N%20Databook%20with%20Copyright.pdf
.. _[BOOK-I]: https://goes.gsfc.nasa.gov/text/databook/databook.pdf
.. _[IR]: https://www.ospo.noaa.gov/Operations/GOES/calibration/gvar-conversion.html
.. _[VIS]: https://www.ospo.noaa.gov/Operations/GOES/calibration/goes-vis-ch-calibration.html
.. _[GLOSSARY]: https://www.avl.class.noaa.gov/release/glossary/GVAR_IMG.htm
.. _[SCHED-W]: https://www.ospo.noaa.gov/Operations/GOES/15/imager-routine.html
.. _[SCHED-E]: http://www.ospo.noaa.gov/Operations/GOES/east/imager-routine.html
"""

import datetime as dt
import logging
import re
from abc import abstractmethod

import numpy as np
import pyresample.geometry
import xarray as xr

from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.utils import bbox, get_geostationary_angle_extent
from satpy.readers.goes_imager_hrit import ALTITUDE, EQUATOR_RADIUS, POLE_RADIUS, SPACECRAFTS
from satpy.utils import get_legacy_chunk_size

logger = logging.getLogger(__name__)

CHUNK_SIZE = get_legacy_chunk_size()

# Radiation constants. Source: [VIS]
C1 = 1.191066E-5  # [mW/(m2-sr-cm-4)]
C2 = 1.438833  # [K/cm-1]

# Calibration Coefficients
#
# VIS Channel
# ============
# slope, offset: Pre-Launch slope & offset for converting counts to radiance
#                (one per detector) [W m-2 um-1 sr-1].
# x0: Space count
# k: pi / (solar spectral irradiance averaged over the spectral response
#          function of the detector) [m2 sr um W-1]
#
#
# IR Channels
# ============
# scale, offset: Scale & offset for converting counts to radiance. Units:
#                [mW m-2 cm-1 sr-1], [1]. They are identical for all platforms.
# n: The channel's central wavenumber (one for each detector) [cm-1]
# a, b: Offset and slope for converting effective BT to actual BT (one per
#       detector). Units: [K], [1]
# btmin, btmax: Valid BT range [K]. Values outside this range will be masked.
#               Extracted from lookup tables provided in [IR].
SCALE_03_9 = 227.3889
OFFSET_03_9 = 68.2167
SCALE_06_8 = 38.8383
OFFSET_06_8 = 29.1287
SCALE_06_5 = 38.8383
OFFSET_06_5 = 29.1287
SCALE_10_7 = 5.2285
OFFSET_10_7 = 15.6854
SCALE_12_0 = 5.0273
OFFSET_12_0 = 15.3332
SCALE_13_3 = 5.5297
OFFSET_13_3 = 16.5892
CALIB_COEFS = {
    "GOES-15": {"00_7": {"slope": [5.851966E-1, 5.879772E-1, 5.856793E-1,
                                   5.854250E-1, 5.866992E-1, 5.836241E-1,
                                   5.846555E-1, 5.843753E-1],
                         "offset": [-16.9707, -17.0513, -16.9847, -16.9773,
                                    -17.0143, -16.9251, -16.9550, -16.9469],
                         "x0": 29,
                         "k": 1.88852E-3},
                "03_9": {"scale": SCALE_03_9,
                         "offset": OFFSET_03_9,
                         "n": [2562.7905, 2562.7905],
                         "a": [-1.5693377, -1.5693377],
                         "b": [1.0025034, 1.0025034],
                         "btmin": 205.0,
                         "btmax": 340.0},
                "06_5": {"scale": SCALE_06_8,
                         "offset": OFFSET_06_8,
                         "n": [1521.1988, 1521.5277],
                         "a": [-3.4706545, -3.4755568],
                         "b": [1.0093296, 1.0092838],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "10_7": {"scale": SCALE_10_7,
                         "offset": OFFSET_10_7,
                         "n": [935.89417, 935.78158],
                         "a": [-0.36151367, -0.35316361],
                         "b": [1.0012715, 1.0012570],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "13_3": {"scale": SCALE_13_3,
                         "offset": OFFSET_13_3,
                         "n": [753.72229, 753.93403],
                         "a": [-0.21475817, -0.24630068],
                         "b": [1.0006485, 1.0007178],
                         "btmin": 180.0,
                         "btmax": 340.0}
                },  # ITT RevH + STAR Correction
    "GOES-14": {"00_7": {"slope": [5.874693E-1, 5.865367E-1, 5.862807E-1,
                                   5.864086E-1, 5.857146E-1, 5.852004E-1,
                                   5.860814E-1, 5.841697E-1],
                         "offset": [-17.037, -17.010, -17.002, -17.006,
                                    -16.986, -16.971, -16.996, -16.941],
                         "x0": 29,
                         "k": 1.88772E-3},
                "03_9": {"scale": SCALE_03_9,
                         "offset": OFFSET_03_9,
                         "n": [2577.3518, 2577.3518],
                         "a": [-1.5297091, -1.5297091],
                         "b": [1.0025608, 1.0025608],
                         "btmin": 205.0,
                         "btmax": 340.0},
                "06_5": {"scale": SCALE_06_8,
                         "offset": OFFSET_06_8,
                         "n": [1519.3488, 1518.5610],
                         "a": [-3.4647892, -3.4390527],
                         "b": [1.0093656, 1.0094427],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "10_7": {"scale": SCALE_10_7,
                         "offset": OFFSET_10_7,
                         "n": [933.98541, 934.19579],
                         "a": [-0.29201763, -0.31824779],
                         "b": [1.0012018, 1.0012303],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "13_3": {"scale": SCALE_13_3,
                         "offset": OFFSET_13_3,
                         "n": [752.88143, 752.82392],
                         "a": [-0.22508805, -0.21700982],
                         "b": [1.0006686, 1.0006503],
                         "btmin": 180.0,
                         "btmax": 340.0}
                },  # ITT RevH + STAR Correction
    "GOES-13": {"00_7": {"slope": [6.120196E-1, 6.118504E-1, 6.096360E-1,
                                   6.087055E-1, 6.132860E-1, 6.118208E-1,
                                   6.122307E-1, 6.066968E-1],
                         "offset": [-17.749, -17.744, -17.769, -17.653,
                                    -17.785, -17.743, -17.755, -17.594],
                         "x0": 29,
                         "k": 1.89544E-3},
                "03_9": {"scale": SCALE_03_9,
                         "offset": OFFSET_03_9,
                         "n": [2561.74, 2561.74],
                         "a": [-1.437204, -1.437204],
                         "b": [1.002562, 1.002562],
                         "btmin": 205.0,
                         "btmax": 340.0},
                "06_5": {"scale": SCALE_06_8,
                         "offset": OFFSET_06_8,
                         "n": [1522.52, 1521.66],
                         "a": [-3.625663, -3.607841],
                         "b": [1.010018, 1.010010],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "10_7": {"scale": SCALE_10_7,
                         "offset": OFFSET_10_7,
                         "n": [937.23, 937.27],
                         "a": [-0.386043, -0.380113],
                         "b": [1.001298, 1.001285],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "13_3": {"scale": SCALE_13_3,
                         "offset": OFFSET_13_3,
                         "n": [749.83],
                         "a": [-0.134801],
                         "b": [1.000482],
                         "btmin": 180.0,
                         "btmax": 340.0}  # Has only one detector on GOES-13
                },
    "GOES-12": {"00_7": {"slope": [5.771030E-1, 5.761764E-1, 5.775825E-1,
                                   5.790699E-1, 5.787051E-1, 5.755969E-1,
                                   5.753973E-1, 5.752099E-1],
                         "offset": [-16.736, -16.709, -16.750, -16.793,
                                    -16.782, -16.692, -16.687, -16.681],
                         "x0": 29,
                         "k": 1.97658E-3},
                "03_9": {"scale": SCALE_03_9,
                         "offset": OFFSET_03_9,
                         "n": [2562.45, 2562.45],
                         "a": [-0.650731, -0.650731],
                         "b": [1.001520, 1.001520],
                         "btmin": 205.0,
                         "btmax": 340.0},
                "06_5": {"scale": SCALE_06_8,
                         "offset": OFFSET_06_8,
                         "n": [1536.43, 1536.94],
                         "a": [-4.764728, -4.775517],
                         "b": [1.012420, 1.012403],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "10_7": {"scale": SCALE_10_7,
                         "offset": OFFSET_10_7,
                         "n": [933.21, 933.21],
                         "a": [-0.360331, -0.360331],
                         "b": [1.001306, 1.001306],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "13_3": {"scale": SCALE_13_3,
                         "offset": OFFSET_13_3,
                         "n": [751.91],
                         "a": [-0.253449],
                         "b": [1.000743],
                         "btmin": 180.0,
                         "btmax": 340.0}  # Has only one detector on GOES-12
                },
    "GOES-11": {"00_7": {"slope": [5.561568E-1, 5.552979E-1, 5.558981E-1,
                                   5.577627E-1, 5.557238E-1, 5.587978E-1,
                                   5.586530E-1, 5.528971E-1],
                         "offset": [-16.129, -16.104, -16.121, -16.175,
                                    -16.116, -16.205, -16.201, -16.034],
                         "x0": 29,
                         "k": 2.01524E-3},
                "03_9": {"scale": SCALE_03_9,
                         "offset": OFFSET_03_9,
                         "n": [2562.07, 2562.07],
                         "a": [-0.644790, -0.644790],
                         "b": [1.000775, 1.000775],
                         "btmin": 205.0,
                         "btmax": 340.0},
                "06_8": {"scale": SCALE_06_8,
                         "offset": OFFSET_06_8,
                         "n": [1481.53],
                         "a": [-0.543401],
                         "b": [1.001495],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "10_7": {"scale": SCALE_10_7,
                         "offset": OFFSET_10_7,
                         "n": [931.76, 931.76],
                         "a": [-0.306809, -0.306809],
                         "b": [1.001274, 1.001274],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "12_0": {"scale": SCALE_12_0,
                         "offset": OFFSET_12_0,
                         "n": [833.67, 833.04],
                         "a": [-0.333216, -0.315110],
                         "b": [1.001000, 1.000967],
                         "btmin": 180.0,
                         "btmax": 340.0}
                },
    "GOES-10": {"00_7": {"slope": [5.605602E-1, 5.563529E-1, 5.566574E-1,
                                   5.582154E-1, 5.583361E-1, 5.571736E-1,
                                   5.563135E-1, 5.613536E-1],
                         "offset": [-16.256, -16.134, -16.143, -16.188,
                                    -16.192, -16.158, -16.133, -16.279],
                         "x0": 29,
                         "k": 1.98808E-3},
                "03_9": {"scale": SCALE_03_9,
                         "offset": OFFSET_03_9,
                         "n": [2552.9845, 2552.9845],
                         "a": [-0.60584483, -0.60584483],
                         "b": [1.0011017, 1.0011017],
                         "btmin": 205.0,
                         "btmax": 340.0},
                "06_8": {"scale": SCALE_06_8,
                         "offset": OFFSET_06_8,
                         "n": [1486.2212],
                         "a": [-0.61653805],
                         "b": [1.0014011],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "10_7": {"scale": SCALE_10_7,
                         "offset": OFFSET_10_7,
                         "n": [936.10260, 935.98981],
                         "a": [-0.27128884, -0.27064036],
                         "b": [1.0009674, 1.0009687],
                         "btmin": 180.0,
                         "btmax": 340.0},
                "12_0": {"scale": SCALE_12_0,
                         "offset": OFFSET_12_0,
                         "n": [830.88473, 830.89691],
                         "a": [-0.26505411, -0.26056452],
                         "b": [1.0009087, 1.0008962],
                         "btmin": 180.0,
                         "btmax": 340.0}
                },
    "GOES-9": {"00_7": {"slope": [0.5492361],
                        "offset": [-15.928],
                        "x0": 29,
                        "k": 1.94180E-3},
               "03_9": {"scale": SCALE_03_9,
                        "offset": OFFSET_03_9,
                        "n": [2555.18, 2555.18],
                        "a": [-0.579908, -0.579908],
                        "b": [1.000942, 1.000942],
                        "btmin": 205.0,
                        "btmax": 340.0},
               "06_8": {"scale": SCALE_06_8,
                        "offset": OFFSET_06_8,
                        "n": [1481.82],
                        "a": [-0.493016],
                        "b": [1.001076],
                        "btmin": 180.0,
                        "btmax": 340.0},
               "10_7": {"scale": SCALE_10_7,
                        "offset": OFFSET_10_7,
                        "n": [934.59, 934.28],
                        "a": [-0.384798, -0.363703],
                        "b": [1.001293, 1.001272],
                        "btmin": 180.0,
                        "btmax": 340.0},
               "12_0": {"scale": SCALE_12_0,
                        "offset": OFFSET_12_0,
                        "n": [834.02, 834.09],
                        "a": [-0.302995, -0.306838],
                        "b": [1.000941, 1.000948],
                        "btmin": 180.0,
                        "btmax": 340.0}
               },
    "GOES-8": {"00_7": {"slope": [0.5501873],
                        "offset": [-15.955],
                        "x0": 29,
                        "k": 1.92979E-3},
               "03_9": {"scale": SCALE_03_9,
                        "offset": OFFSET_03_9,
                        "n": [2556.71, 2558.62],
                        "a": [-0.578526, -0.581853],
                        "b": [1.001512, 1.001532],
                        "btmin": 205.0,
                        "btmax": 340.0},
               "06_8": {"scale": SCALE_06_8,
                        "offset": OFFSET_06_8,
                        "n": [1481.91],
                        "a": [-0.593903],
                        "b": [1.001418],
                        "btmin": 180.0,
                        "btmax": 340.0},
               "10_7": {"scale": SCALE_10_7,
                        "offset": OFFSET_10_7,
                        "n": [934.30, 935.38],
                        "a": [-0.322585, -0.351889],
                        "b": [1.001271, 1.001293],
                        "btmin": 180.0,
                        "btmax": 340.0},
               "12_0": {"scale": SCALE_12_0,
                        "offset": OFFSET_12_0,
                        "n": [837.06, 837.00],
                        "a": [-0.422571, -0.466954],
                        "b": [1.001170, 1.001257],
                        "btmin": 180.0,
                        "btmax": 340.0}
               }
}

# Angular sampling rates in radians. Source: [BOOK-I], [BOOK-N]
SAMPLING_EW_VIS = 16E-6
SAMPLING_NS_VIS = 28E-6
SAMPLING_EW_IR = 64E-6
SAMPLING_NS_IR = 112E-6

# Sector definitions. TODO: Add remaining sectors (PACUS, CONUS, ...)
FULL_DISC = "Full Disc"
NORTH_HEMIS_EAST = "Northern Hemisphere (GOES-East)"
SOUTH_HEMIS_EAST = "Southern Hemisphere (GOES-East)"
NORTH_HEMIS_WEST = "Northern Hemisphere (GOES-West)"
SOUTH_HEMIS_WEST = "Southern Hemisphere (GOES-West)"
UNKNOWN_SECTOR = "Unknown"

IR_SECTORS = {
    (2704, 5208): FULL_DISC,
    (1826, 3464): NORTH_HEMIS_EAST,
    (566, 3464): SOUTH_HEMIS_EAST,
    (1354, 3312): NORTH_HEMIS_WEST,
    (1062, 2760): SOUTH_HEMIS_WEST
}  # (nlines, ncols)

VIS_SECTORS = {
    (10819, 20800): FULL_DISC,
    (7307, 13852): NORTH_HEMIS_EAST,
    (2267, 13852): SOUTH_HEMIS_EAST,
    (5419, 13244): NORTH_HEMIS_WEST,
    (4251, 11044): SOUTH_HEMIS_WEST
}  # (nlines, ncols)

SCAN_DURATION = {
    FULL_DISC: dt.timedelta(minutes=26),
    NORTH_HEMIS_WEST: dt.timedelta(minutes=10, seconds=5),
    SOUTH_HEMIS_WEST: dt.timedelta(minutes=6, seconds=54),
    NORTH_HEMIS_EAST: dt.timedelta(minutes=14, seconds=15),
    SOUTH_HEMIS_EAST: dt.timedelta(minutes=4, seconds=49)
}  # Source: [SCHED-W], [SCHED-E]


class GOESNCBaseFileHandler(BaseFileHandler):
    """File handler for GOES Imager data in netCDF format."""

    yaw_flip_sampling_distance = 10

    def __init__(self, filename, filename_info, filetype_info, geo_data=None):
        """Initialize the reader."""
        super(GOESNCBaseFileHandler, self).__init__(filename, filename_info,
                                                    filetype_info)
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  chunks={"xc": CHUNK_SIZE, "yc": CHUNK_SIZE})
        self.sensor = "goes_imager"
        self.nlines = self.nc.sizes["yc"]
        self.ncols = self.nc.sizes["xc"]
        self.platform_name = self._get_platform_name(
            self.nc.attrs["Satellite Sensor"])
        self.platform_shortname = self.platform_name.replace("-", "").lower()
        self.gvar_channel = int(self.nc["bands"].item())
        self.sector = self._get_sector(channel=self.gvar_channel,
                                       nlines=self.nlines,
                                       ncols=self.ncols)
        self._meta = None
        self.geo_data = geo_data if geo_data is not None else self.nc

    @abstractmethod
    def get_dataset(self, key, info):
        """Load dataset designated by the given key from file."""
        raise NotImplementedError

    @abstractmethod
    def calibrate(self, data, calibration, channel):
        """Perform calibration."""
        raise NotImplementedError

    @property
    @abstractmethod
    def vis_sectors(self):
        """Get the vis sectors."""
        raise NotImplementedError

    @property
    @abstractmethod
    def ir_sectors(self):
        """Get the ir sectors."""
        raise NotImplementedError

    @staticmethod
    def _get_platform_name(ncattr):
        """Determine name of the platform."""
        match = re.match(r"G-(\d+)", ncattr)
        if match:
            return SPACECRAFTS.get(int(match.groups()[0]))

        return None

    def _get_sector(self, channel, nlines, ncols):
        """Determine which sector was scanned."""
        if is_vis_channel(channel):
            margin = 100
            sectors_ref = self.vis_sectors
        else:
            margin = 50
            sectors_ref = self.ir_sectors

        for (nlines_ref, ncols_ref), sector in sectors_ref.items():
            if np.fabs(ncols - ncols_ref) < margin and \
                    np.fabs(nlines - nlines_ref) < margin:
                return sector

        return UNKNOWN_SECTOR

    @staticmethod
    def _get_earth_mask(lat):
        """Identify earth/space pixels.

        Returns:
            Mask (1=earth, 0=space)
        """
        logger.debug("Computing earth mask")
        return np.fabs(lat) <= 90

    @staticmethod
    def _get_nadir_pixel(earth_mask, sector):
        """Find the nadir pixel.

        Args:
            earth_mask: Mask identifying earth and space pixels
            sector: Specifies the scanned sector
        Returns:
            nadir row, nadir column
        """
        if sector == FULL_DISC:
            logger.debug("Computing nadir pixel")

            # The earth is not centered in the image, compute bounding box
            # of the earth disc first
            rmin, rmax, cmin, cmax = bbox(earth_mask)

            # The nadir pixel is approximately at the centre of the earth disk
            nadir_row = rmin + (rmax - rmin) // 2
            nadir_col = cmin + (cmax - cmin) // 2

            return nadir_row, nadir_col

        return None, None

    def _is_yaw_flip(self, lat):
        """Determine whether the satellite is yaw-flipped ('upside down')."""
        logger.debug("Computing yaw flip flag")
        # In case of yaw-flip the data and coordinates in the netCDF files are
        # also flipped. Just check whether the latitude increases or decrases
        # with the line number.
        delta = self.yaw_flip_sampling_distance
        crow, ccol = np.array(lat.shape) // 2
        return (lat[crow+delta, ccol] - lat[crow, ccol]).values > 0

    def _get_area_def_uniform_sampling(self, lon0, channel):
        """Get area definition with uniform sampling."""
        logger.debug("Computing area definition")
        if lon0 is not None:
            est = AreaDefEstimator(self.platform_name, channel)
            return est.get_area_def_with_uniform_sampling(lon0)
        return None

    @property
    def start_time(self):
        """Start timestamp of the dataset."""
        timestamp = self.nc["time"].dt
        return dt.datetime(
            year=int(timestamp.year.item()),
            month=int(timestamp.month.item()),
            day=int(timestamp.day.item()),
            hour=int(timestamp.hour.item()),
            minute=int(timestamp.minute.item()),
            second=int(timestamp.second.item()),
            microsecond=int(timestamp.microsecond.item()))

    @property
    def end_time(self):
        """End timestamp of the dataset."""
        try:
            return self.start_time + SCAN_DURATION[self.sector]
        except KeyError:
            return self.start_time

    @property
    def resolution(self):
        """Specify the spatial resolution of the dataset.

        Channel 13_3's spatial resolution changes from one platform to another
        while the wavelength and file format remain the same. In order to
        avoid multiple YAML reader definitions for the same file format,
        read the channel's resolution from the file instead of defining it
        in the YAML dataset. This information will then be used by the YAML
        reader to complement the YAML definition of the dataset.

        Returns:
            Spatial resolution in kilometers
        """
        return 1000. * self.nc["lineRes"].values

    def get_shape(self, key, info):
        """Get the shape of the data.

        Returns:
            Number of lines, number of columns
        """
        return self.nlines, self.ncols

    @property
    def meta(self):
        """Derive metadata from the coordinates."""
        # Use buffered data if available
        if self._meta is None:
            lat = self.geo_data["lat"]
            earth_mask = self._get_earth_mask(lat)
            crow, ccol = self._get_nadir_pixel(earth_mask=earth_mask,
                                               sector=self.sector)
            lat0 = lat.values[crow, ccol] if crow is not None else None
            yaw_flip = self._is_yaw_flip(lat)
            del lat

            lon = self.geo_data["lon"]
            lon0 = lon.values[crow, ccol] if crow is not None else None
            area_def_uni = self._get_area_def_uniform_sampling(
                lon0=lon0, channel=self.gvar_channel)
            del lon

            self._meta = {"earth_mask": earth_mask,
                          "yaw_flip": yaw_flip,
                          "lat0": lat0,
                          "lon0": lon0,
                          "nadir_row": crow,
                          "nadir_col": ccol,
                          "area_def_uni": area_def_uni}
        return self._meta

    def _counts2radiance(self, counts, coefs, channel):
        """Convert raw detector counts to radiance."""
        logger.debug("Converting counts to radiance")

        if is_vis_channel(channel):
            # Since the scanline-detector assignment is unknown, use the average
            # coefficients for all scanlines.
            slope = np.array(coefs["slope"]).mean()
            offset = np.array(coefs["offset"]).mean()
            return self._viscounts2radiance(counts=counts, slope=slope,
                                            offset=offset)

        return self._ircounts2radiance(counts=counts, scale=coefs["scale"],
                                       offset=coefs["offset"])

    def _calibrate(self, radiance, coefs, channel, calibration):
        """Convert radiance to reflectance or brightness temperature."""
        if is_vis_channel(channel):
            if not calibration == "reflectance":
                raise ValueError("Cannot calibrate VIS channel to "
                                 "{}".format(calibration))
            return self._calibrate_vis(radiance=radiance, k=coefs["k"])
        else:
            if not calibration == "brightness_temperature":
                raise ValueError("Cannot calibrate IR channel to "
                                 "{}".format(calibration))

            # Since the scanline-detector assignment is unknown, use the average
            # coefficients for all scanlines.
            mean_coefs = {"a": np.array(coefs["a"]).mean(),
                          "b": np.array(coefs["b"]).mean(),
                          "n": np.array(coefs["n"]).mean(),
                          "btmin": coefs["btmin"],
                          "btmax": coefs["btmax"]}
            return self._calibrate_ir(radiance=radiance, coefs=mean_coefs)

    @staticmethod
    def _ircounts2radiance(counts, scale, offset):
        """Convert IR counts to radiance.

        Reference: [IR].

        Args:
            counts: Raw detector counts
            scale: Scale [mW-1 m2 cm sr]
            offset: Offset [1]

        Returns:
            Radiance [mW m-2 cm-1 sr-1]
        """
        rad = (counts - offset) / scale
        return rad.clip(min=0)

    @staticmethod
    def _calibrate_ir(radiance, coefs):
        """Convert IR radiance to brightness temperature.

        Reference: [IR]

        Args:
            radiance: Radiance [mW m-2 cm-1 sr-1]
            coefs: Dictionary of calibration coefficients. Keys:
                   n: The channel's central wavenumber [cm-1]
                   a: Offset [K]
                   b: Slope [1]
                   btmin: Minimum brightness temperature threshold [K]
                   btmax: Maximum brightness temperature threshold [K]

        Returns:
            Brightness temperature [K]
        """
        logger.debug("Calibrating to brightness temperature")

        # Compute brightness temperature using inverse Planck formula
        n = coefs["n"]
        bteff = C2 * n / np.log(1 + C1 * n ** 3 / radiance.where(radiance > 0))
        bt = xr.DataArray(bteff * coefs["b"] + coefs["a"])

        # Apply BT threshold
        return bt.where(np.logical_and(bt >= coefs["btmin"],
                                       bt <= coefs["btmax"]))

    @staticmethod
    def _viscounts2radiance(counts, slope, offset):
        """Convert VIS counts to radiance.

        References: [VIS]

        Args:
            counts: Raw detector counts
            slope: Slope [W m-2 um-1 sr-1]
            offset: Offset [W m-2 um-1 sr-1]

        Returns:
            Radiance [W m-2 um-1 sr-1]
        """
        rad = counts * slope + offset
        return rad.clip(min=0)

    @staticmethod
    def _calibrate_vis(radiance, k):
        """Convert VIS radiance to reflectance.

        Note: Angle of incident radiation and annual variation of the
        earth-sun distance is not taken into account. A value of 100%
        corresponds to the radiance of a perfectly reflecting diffuse surface
        illuminated at normal incidence when the sun is at its annual-average
        distance from the Earth.

        TODO: Take angle of incident radiation (cos sza) and annual variation
        of the earth-sun distance into account.

        Reference: [VIS]

        Args:
            radiance: Radiance [mW m-2 cm-1 sr-1]
            k: pi / H, where H is the solar spectral irradiance at
               annual-average sun-earth distance, averaged over the spectral
               response function of the detector). Units of k: [m2 um sr W-1]

        Returns:
            Reflectance [%]
        """
        logger.debug("Calibrating to reflectance")
        refl = 100 * k * radiance
        return refl.clip(min=0)

    def _update_metadata(self, data, ds_info):
        """Update metadata of the given DataArray."""
        # Metadata from the dataset definition
        data.attrs.update(ds_info)

        # If the file_type attribute is a list and the data is xarray
        # the concat of the dataset will not work. As the file_type is
        # not needed this will be popped here.
        if "file_type" in data.attrs:
            data.attrs.pop("file_type")

        # Metadata discovered from the file.
        data.attrs.update(
            {"platform_name": self.platform_name,
             "sensor": self.sensor,
             "sector": self.sector,
             "orbital_parameters": {"yaw_flip": self.meta["yaw_flip"]}}
        )
        if self.meta["lon0"] is not None:
            # Attributes only available for full disc images. YAML reader
            # doesn't like it if satellite_* is present but None
            data.attrs.update(
                {"nadir_row": self.meta["nadir_row"],
                 "nadir_col": self.meta["nadir_col"],
                 "area_def_uniform_sampling": self.meta["area_def_uni"]}
            )
            data.attrs["orbital_parameters"].update(
                {"projection_longitude": self.meta["lon0"],
                 "projection_latitude": self.meta["lat0"],
                 "projection_altitude": ALTITUDE}
            )

    def __del__(self):
        """Delete."""
        try:
            self.nc.close()
        except (AttributeError, OSError):
            pass

    def available_datasets(self, configured_datasets=None):
        """Update information for or add datasets provided by this file.

        If this file handler can load a dataset then it will supplement the
        dataset info with the resolution and possibly coordinate datasets
        needed to load it. Otherwise it will continue passing the dataset
        information down the chain.

        See
        :meth:`satpy.readers.core.file_handlers.BaseFileHandler.available_datasets`
        for details.
        """
        res = self.resolution
        # update previously configured datasets
        for is_avail, ds_info in (configured_datasets or []):
            if is_avail is not None:
                yield is_avail, ds_info

            matches = self.file_type_matches(ds_info["file_type"])
            if matches and ds_info.get("resolution") != res:
                new_info = ds_info.copy()
                new_info["resolution"] = res
                yield True, new_info
            elif is_avail is None:
                yield is_avail, ds_info


def is_vis_channel(channel):
    """Determine whether the given channel is a visible channel."""
    if isinstance(channel, str):
        return channel == "00_7"
    if isinstance(channel, int):
        return channel == 1
    raise ValueError("Invalid channel")


class GOESNCFileHandler(GOESNCBaseFileHandler):
    """File handler for GOES Imager data in netCDF format."""

    vis_sectors = VIS_SECTORS
    ir_sectors = IR_SECTORS

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(GOESNCFileHandler, self).__init__(filename, filename_info,
                                                filetype_info)

    def get_dataset(self, key, info):
        """Load dataset designated by the given key from file."""
        logger.debug("Reading dataset {}".format(key["name"]))

        # Read data from file and calibrate if necessary
        if "longitude" in key["name"]:
            data = self.geo_data["lon"]
        elif "latitude" in key["name"]:
            data = self.geo_data["lat"]
        else:
            tic = dt.datetime.now()
            data = self.calibrate(self.nc["data"].isel(time=0),
                                  calibration=key["calibration"],
                                  channel=key["name"])
            logger.debug("Calibration time: {}".format(dt.datetime.now() - tic))

        # Mask space pixels
        data = data.where(self.meta["earth_mask"])

        # Set proper dimension names
        data = data.rename({"xc": "x", "yc": "y"})

        # Update metadata
        self._update_metadata(data, ds_info=info)

        return data

    def calibrate(self, counts, calibration, channel):
        """Perform calibration."""
        # Convert 16bit counts from netCDF4 file to the original 10bit
        # GVAR counts by dividing by 32. See [GLOSSARY].
        counts = counts / 32.

        coefs = CALIB_COEFS[self.platform_name][channel]
        if calibration == "counts":
            return counts
        if calibration in ["radiance", "reflectance",
                           "brightness_temperature"]:
            radiance = self._counts2radiance(counts=counts, coefs=coefs,
                                             channel=channel)
            if calibration == "radiance":
                return radiance

            return self._calibrate(radiance=radiance, coefs=coefs,
                                   channel=channel, calibration=calibration)

        raise ValueError("Unsupported calibration for channel {}: {}".format(channel, calibration))


class GOESEUMNCFileHandler(GOESNCBaseFileHandler):
    """File handler for GOES Imager data in EUM netCDF format.

    TODO: Remove datasets which are not available in the file (counts,
    VIS radiance) via available_datasets()  -> See #434
    """

    vis_sectors = IR_SECTORS  # VIS channel is downsampled to IR resolution
    ir_sectors = IR_SECTORS

    def __init__(self, filename, filename_info, filetype_info, geo_data):
        """Initialize the reader."""
        super(GOESEUMNCFileHandler, self).__init__(filename, filename_info,
                                                   filetype_info, geo_data)

    def get_dataset(self, key, info):
        """Load dataset designated by the given key from file."""
        logger.debug("Reading dataset {}".format(key["name"]))

        tic = dt.datetime.now()
        data = self.calibrate(self.nc["data"].isel(time=0),
                              calibration=key["calibration"],
                              channel=key["name"])
        logger.debug("Calibration time: {}".format(dt.datetime.now() - tic))

        # Mask space pixels
        data = data.where(self.meta["earth_mask"])

        # Set proper dimension names
        data = data.rename({"xc": "x", "yc": "y"})
        data = data.drop_vars("time")

        # Update metadata
        self._update_metadata(data, ds_info=info)

        return data

    def calibrate(self, data, calibration, channel):
        """Perform calibration."""
        coefs = CALIB_COEFS[self.platform_name][channel]
        is_vis = is_vis_channel(channel)

        # IR files provide radiances, VIS file provides reflectances
        if is_vis and calibration == "reflectance":
            return data
        if not is_vis and calibration == "radiance":
            return data
        if not is_vis and calibration == "brightness_temperature":
            return self._calibrate(radiance=data, calibration=calibration,
                                   coefs=coefs, channel=channel)

        raise ValueError("Unsupported calibration for channel {}: {}"
                         .format(channel, calibration))


class GOESEUMGEONCFileHandler(BaseFileHandler):
    """File handler for GOES Geolocation data in EUM netCDF format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Initialize the reader."""
        super(GOESEUMGEONCFileHandler, self).__init__(filename, filename_info,
                                                      filetype_info)
        self.nc = xr.open_dataset(self.filename,
                                  decode_cf=True,
                                  mask_and_scale=False,
                                  chunks={"xc": CHUNK_SIZE, "yc": CHUNK_SIZE})
        self.sensor = "goes_imager"
        self.nlines = self.nc.sizes["yc"]
        self.ncols = self.nc.sizes["xc"]
        self.platform_name = GOESNCBaseFileHandler._get_platform_name(
            self.nc.attrs["Satellite Sensor"])
        self.platform_shortname = self.platform_name.replace("-", "").lower()
        self._meta = None

    def __getitem__(self, item):
        """Get item."""
        return getattr(self.nc, item)

    def get_dataset(self, key, info):
        """Load dataset designated by the given key from file."""
        logger.debug("Reading dataset {}".format(key["name"]))

        # Read data from file and calibrate if necessary
        if "longitude" in key["name"]:
            data = self.nc["lon"]
        elif "latitude" in key["name"]:
            data = self.nc["lat"]
        else:
            raise KeyError("Unknown dataset: {}".format(key["name"]))

        # Set proper dimension names
        data = data.rename({"xc": "x", "yc": "y"})

        # Update metadata
        data.attrs.update(info)
        return data

    @property
    def resolution(self):
        """Specify the spatial resolution of the dataset.

        In the EUMETSAT format VIS data is downsampled to IR resolution (4km).
        """
        return 4000.0


class GOESCoefficientReader(object):
    """Read GOES Imager calibration coefficients from NOAA reference HTMLs."""

    gvar_channels = {
        "GOES-8": {"00_7": 1, "03_9": 2, "06_8": 3, "10_7": 4, "12_0": 5},
        "GOES-9": {"00_7": 1, "03_9": 2, "06_8": 3, "10_7": 4, "12_0": 5},
        "GOES-10": {"00_7": 1, "03_9": 2, "06_8": 3, "10_7": 4, "12_0": 5},
        "GOES-11": {"00_7": 1, "03_9": 2, "06_8": 3, "10_7": 4, "12_0": 5},
        "GOES-12": {"00_7": 1, "03_9": 2, "06_5": 3, "10_7": 4, "13_3": 6},
        "GOES-13": {"00_7": 1, "03_9": 2, "06_5": 3, "10_7": 4, "13_3": 6},
        "GOES-14": {"00_7": 1, "03_9": 2, "06_5": 3, "10_7": 4, "13_3": 6},
        "GOES-15": {"00_7": 1, "03_9": 2, "06_5": 3, "10_7": 4, "13_3": 6},
    }

    ir_tables = {
        "GOES-8": "2-1",
        "GOES-9": "2-2",
        "GOES-10": "2-3",
        "GOES-11": "2-4",
        "GOES-12": "2-5a",
        "GOES-13": "2-6",
        "GOES-14": "2-7c",
        "GOES-15": "2-8b"
    }

    vis_tables = {
        "GOES-8": "Table 1.",
        "GOES-9": "Table 1.",
        "GOES-10": "Table 2.",
        "GOES-11": "Table 3.",
        "GOES-12": "Table 4.",
        "GOES-13": "Table 5.",
        "GOES-14": "Table 6.",
        "GOES-15": "Table 7."
    }

    def __init__(self, ir_url, vis_url):
        """Init the coef reader."""
        from bs4 import BeautifulSoup
        self.ir_html = BeautifulSoup(self._load_url_or_file(ir_url),
                                     features="html5lib")
        self.vis_html = BeautifulSoup(self._load_url_or_file(vis_url),
                                      features="html5lib")

    def _load_url_or_file(self, url):
        import requests
        from requests.exceptions import MissingSchema

        try:
            response = requests.get(url, timeout=60)
            if response.ok:
                return response.text
            raise requests.HTTPError
        except (MissingSchema, requests.HTTPError):
            # Not a valid URL, is it a file?
            try:
                return open(url, mode="r")
            except IOError:
                raise ValueError("Invalid URL or file: {}".format(url))

    def get_coefs(self, platform, channel):
        """Get the coefs."""
        if channel == "00_7":
            return self._get_vis_coefs(platform=platform)

        return self._get_ir_coefs(platform=platform, channel=channel)

    def _get_ir_coefs(self, platform, channel):
        from collections import defaultdict

        coefs = defaultdict(list)

        # Extract scale and offset for conversion counts->radiance from
        # Table 1-1 (same for all platforms, only depends on the channel)
        gvar_channel = self.gvar_channels[platform][channel]
        table11 = self._get_table(root=self.ir_html, heading="Table 1-1",
                                  heading_type="h3")
        for row in table11:
            if int(row[0]) == gvar_channel:
                coefs["scale"] = self._float(row[1])
                coefs["offset"] = self._float(row[2])

        # Extract n,a,b (radiance -> BT) from the coefficient table for the
        # given platform
        table = self._get_table(root=self.ir_html,
                                heading=self.ir_tables[platform],
                                heading_type="h3")
        channel_regex = re.compile("^{}(?:/[a,b])?$".format(gvar_channel))

        for row in table:
            if channel_regex.match(row[0]):
                # Extract coefficients. Detector (a) always comes before (b)
                # in the table so that simply appending preserves the order.
                coefs["n"].append(self._float(row[1]))
                coefs["a"].append(self._float(row[2]))
                coefs["b"].append(self._float(row[3]))

        return coefs

    def _get_vis_coefs(self, platform):
        from collections import defaultdict

        # Find calibration table
        table = self._get_table(root=self.vis_html,
                                heading=self.vis_tables[platform],
                                heading_type="p")

        # Extract values
        coefs = defaultdict(list)
        if platform in ("GOES-8", "GOES-9"):
            # GOES 8&9 coefficients are in the same table
            col = 1 if platform == "GOES-8" else 2
            coefs["slope"].append(self._float(table[1][col]))
            coefs["x0"] = self._float(table[2][col])
            coefs["offset"].append(self._float(table[3][col]))
            coefs["k"] = self._float(table[4][col])
        else:
            # k and x0 appear in the first row only
            coefs["slope"].append(self._float(table[0][1]))
            coefs["x0"] = self._float(table[0][2])
            coefs["k"] = self._float(table[0][4])
            coefs["offset"].append(self._float(table[0][3]))

            # Remaining rows
            for row in table[1:]:
                coefs["slope"].append(self._float(row[1]))
                coefs["offset"].append(self._float(row[2]))

        return coefs

    def _get_table(self, root, heading, heading_type, ):
        # Find table by its heading
        headings = [h for h in root.find_all(heading_type)
                    if heading in h.text]
        if not headings:
            raise ValueError("Cannot find a coefficient table matching text "
                             '"{}"'.format(heading))
        if len(headings) > 1:
            raise ValueError('Found multiple headings matching text "{}"'
                             .format(heading))
        table = headings[0].next_sibling.next_sibling

        # Copy items to a list of lists
        tab = list()
        for row in table.find_all("tr"):
            cols = row.find_all("td")
            if cols:
                tab.append([c.text for c in cols])
        return tab

    def _denoise(self, string):
        return string.replace("\n", "").replace(" ", "")

    def _float(self, string):
        """Convert string to float.

        Take care of numbers in exponential format
        """
        string = self._denoise(string)
        exp_match = re.match(r"^[-.\d]+x10-(\d)$", string)
        if exp_match:
            exp = int(exp_match.groups()[0])
            fac = 10 ** -exp
            string = string.replace("x10-{}".format(exp), "")
        else:
            fac = 1

        return fac * float(string)


def test_coefs(ir_url, vis_url):
    """Test calibration coefficients against NOAA reference pages.

    Currently the reference pages are:

    ir_url = https://www.ospo.noaa.gov/Operations/GOES/calibration/gvar-conversion.html
    vis_url = https://www.ospo.noaa.gov/Operations/GOES/calibration/goes-vis-ch-calibration.html

    Args:
        ir_url: Path or URL to HTML page with IR coefficients
        vis_url: Path or URL to HTML page with VIS coefficients

    Raises:
        (Exception) ValueError: if coefficients don't match the reference
    """
    reader = GOESCoefficientReader(ir_url=ir_url, vis_url=vis_url)

    for platform in CALIB_COEFS:
        for channel, coefs in CALIB_COEFS[platform].items():
            coefs_expected = reader.get_coefs(platform=platform,
                                              channel=channel)
            for cname in coefs_expected.keys():
                if not np.allclose(coefs[cname], coefs_expected[cname]):
                    raise ValueError(
                        "Coefficient {} for {} channel {} does not match the "
                        "reference".format(cname, platform, channel))

    logger.info("Coefficients OK")
    return True


class AreaDefEstimator:
    """Estimate area definition for GOES-Imager."""

    def __init__(self, platform_name, channel):
        """Create the instance."""
        self.platform_name = platform_name
        self.channel = channel

    def get_area_def_with_uniform_sampling(self, projection_longitude):
        """Get area definition with uniform sampling.

        The area definition is based on geometry and instrument properties:
        Pixel size is derived from altitude and N-S sampling angle. Area extent
        is based on the maximum scanning angles at the limb of the earth.
        """
        projection = self._get_projection(projection_longitude)
        area_extent = self._get_area_extent_at_max_scan_angle(projection)
        shape = self._get_shape_with_uniform_pixel_size(area_extent)
        return self._create_area_def(projection, area_extent, shape)

    def _get_projection(self, projection_longitude):
        return {
            "a": EQUATOR_RADIUS,
            "b": POLE_RADIUS,
            "lon_0": projection_longitude,
            "h": ALTITUDE,
            "proj": "geos",
            "units": "m"
        }

    def _get_area_extent_at_max_scan_angle(self, proj_dict):
        xmax, ymax = self._get_max_scan_angle(proj_dict)
        return ALTITUDE * np.array([-xmax, -ymax, xmax, ymax])

    def _get_max_scan_angle(self, proj_dict):
        dummy_area = pyresample.geometry.AreaDefinition(
            area_id="dummy",
            proj_id="dummy",
            description="dummy",
            projection=proj_dict,
            width=2,
            height=2,
            area_extent=[-1, -1, 1, 1]
        )  # only projection is relevant here
        xmax, ymax = get_geostationary_angle_extent(dummy_area)
        return xmax, ymax

    def _get_shape_with_uniform_pixel_size(self, area_extent):
        llx, lly, urx, ury = area_extent
        pix_size = self._get_uniform_pixel_size()
        width = np.rint((urx - llx) / pix_size).astype(int)
        height = np.rint((ury - lly) / pix_size).astype(int)
        return width, height

    def _get_uniform_pixel_size(self):
        if is_vis_channel(self.channel):
            sampling = SAMPLING_NS_VIS
        else:
            sampling = SAMPLING_NS_IR
        pix_size = ALTITUDE * sampling
        return pix_size

    def _create_area_def(self, projection, area_extent, shape):
        width, height = shape
        return pyresample.geometry.AreaDefinition(
            area_id="goes_geos_uniform",
            proj_id="goes_geos_uniform",
            description=self._get_area_description(),
            projection=projection,
            width=width,
            height=height,
            area_extent=area_extent
        )

    def _get_area_description(self):
        return "{} geostationary projection (uniform sampling)".format(
            self.platform_name
        )



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for enhancement testing."""

import dask.array as da
import numpy as np
import xarray as xr

from satpy.tests.utils import assert_maximum_dask_computes


def run_and_check_enhancement(func, data, expected, **kwargs):
    """Perform basic checks that apply to multiple tests."""
    pre_attrs = data.attrs
    with assert_maximum_dask_computes(max_computes=0):
        img = _get_enhanced_image(func, data, **kwargs)

    _assert_image(img, pre_attrs, func.__name__, "palettes" in kwargs)
    _assert_image_data(img, expected)


def _get_enhanced_image(func, data, **kwargs):
    from trollimage.xrimage import XRImage

    img = XRImage(data)
    func(img, **kwargs)

    return img


def _assert_image(img, pre_attrs, func_name, has_palette):
    assert isinstance(img.data, xr.DataArray)
    assert isinstance(img.data.data, da.Array)

    old_keys = set(pre_attrs.keys())
    # It is OK to have "enhancement_history" added
    new_keys = set(img.data.attrs.keys()) - {"enhancement_history"}
    # In case of palettes are used, _FillValue is added.
    # Colorize doesn't add the fill value, so ignore that
    if has_palette and func_name != "colorize":
        assert "_FillValue" in new_keys
        # Remove it from further comparisons
        new_keys = new_keys - {"_FillValue"}
    assert old_keys == new_keys


def _assert_image_data(img, expected, dtype=None):
    # Compute the data to mimic what xrimage geotiff writing does
    res_data = img.data.data.compute()
    assert not isinstance(res_data, da.Array)
    np.testing.assert_allclose(res_data, expected, atol=1.e-6, rtol=0)
    if dtype:
        assert img.data.dtype == dtype
        assert res_data.dtype == dtype


def run_and_check_enhancement_with_dtype(func, data, expected, **kwargs):
    """Perform basic checks that apply to multiple tests."""
    pre_attrs = data.attrs
    img = _get_enhanced_image(func, data, **kwargs)

    _assert_image(img, pre_attrs, func.__name__, "palettes" in kwargs)
    _assert_image_data(img, expected, dtype=data.dtype)


def _create_data():
    data = np.arange(-210, 790, 100).reshape((2, 5)) * 0.95
    data[0, 0] = np.nan  # one bad value for testing

    return data


def create_ch1():
    """Create single channel data."""
    data = _create_data()
    return xr.DataArray(da.from_array(data, chunks=2), dims=("y", "x"), attrs={"test": "test"})


def _create_crefl_data():
    crefl_data = np.arange(-210, 790, 100).reshape((2, 5)) * 0.95
    crefl_data /= 5.605
    crefl_data[0, 0] = np.nan  # one bad value for testing
    crefl_data[0, 1] = 0.

    return crefl_data


def create_ch2():
    """Create test data for crefl."""
    crefl_data = _create_crefl_data()
    return xr.DataArray(da.from_array(crefl_data, chunks=2), dims=("y", "x"), attrs={"test": "test"})


def create_rgb():
    """Create RGB test data."""
    data = _create_data()
    rgb_data = np.stack([data, data, data])
    return xr.DataArray(da.from_array(rgb_data, chunks=(3, 2, 2)),
                        dims=("bands", "y", "x"),
                                coords={"bands": ["R", "G", "B"]})



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2021 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reading and calibrating hrpt avhrr data.

Reading:
http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c4/sec4-1.htm#t413-1

Calibration:
http://www.ncdc.noaa.gov/oa/pod-guide/ncdc/docs/klm/html/c7/sec7-1.htm

"""

import datetime as dt
import logging

import dask.array as da
import numpy as np
import xarray as xr
from geotiepoints import SatelliteInterpolator
from pyorbital.geoloc import compute_pixels, get_lonlatalt
from pyorbital.geoloc_instrument_definitions import avhrr
from pyorbital.orbital import Orbital

from satpy._compat import cached_property
from satpy.readers.aapp_l1b import get_avhrr_lac_chunks
from satpy.readers.core.file_handlers import BaseFileHandler

logger = logging.getLogger(__name__)

AVHRR_CHANNEL_NAMES = ("1", "2", "3a", "3b", "4", "5")

scanline_dtype = np.dtype([("frame_sync", ">u2", (6, )),
                           ("id", [("id", ">u2"),
                                   ("spare", ">u2")]),
                           ("timecode", ">u2", (4, )),
                           ("telemetry", [("ramp_calibration", ">u2", (5, )),
                                          ("PRT", ">u2", (3, )),
                                          ("ch3_patch_temp", ">u2"),
                                          ("spare", ">u2"), ]),
                           ("back_scan", ">u2", (10, 3)),
                           ("space_data", ">u2", (10, 5)),
                           ("sync", ">u2"),
                           ("TIP_data", ">u2", (520, )),
                           ("spare", ">u2", (127, )),
                           ("image_data", ">u2", (2048, 5)),
                           ("aux_sync", ">u2", (100, ))])


def time_seconds(tc_array, year):
    """Return the time object from the timecodes."""
    tc_array = np.array(tc_array, copy=True)
    word = tc_array[:, 0]
    day = word >> 1
    word = tc_array[:, 1].astype(np.uint64)
    msecs = ((127) & word) * 1024
    word = tc_array[:, 2]
    msecs += word & 1023
    msecs *= 1024
    word = tc_array[:, 3]
    msecs += word & 1023
    return (np.datetime64(
        str(year) + "-01-01T00:00:00", "s") +
        msecs[:].astype("timedelta64[ms]") +
        (day - 1)[:].astype("timedelta64[D]"))


def bfield(array, bit):
    """Return the bit array."""
    return (array & 2**(9 - bit + 1)).astype(bool)


spacecrafts = {7: "NOAA 15", 3: "NOAA 16", 13: "NOAA 18", 15: "NOAA 19"}


def geo_interpolate(lons32km, lats32km):
    """Interpolate geo data."""
    cols32km = np.arange(0, 2048, 32)
    cols1km = np.arange(2048)
    lines = lons32km.shape[0]
    rows32km = np.arange(lines)
    rows1km = np.arange(lines)

    along_track_order = 1
    cross_track_order = 3

    satint = SatelliteInterpolator(
        (lons32km, lats32km), (rows32km, cols32km), (rows1km, cols1km),
        along_track_order, cross_track_order)
    lons, lats = satint.interpolate()
    return lons, lats


def _get_channel_index(key):
    """Get the avhrr channel index."""
    avhrr_channel_index = {"1": 0,
                           "2": 1,
                           "3a": 2,
                           "3b": 2,
                           "4": 3,
                           "5": 4}
    index = avhrr_channel_index[key["name"]]
    return index


class HRPTFile(BaseFileHandler):
    """Reader for HRPT Minor Frame, 10 bits data expanded to 16 bits."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(HRPTFile, self).__init__(filename, filename_info, filetype_info)
        self.channels = {i: None for i in AVHRR_CHANNEL_NAMES}
        self.units = {i: "counts" for i in AVHRR_CHANNEL_NAMES}

        self.year = filename_info["start_time"].year

    @cached_property
    def times(self):
        """Get the timestamps for each line."""
        return time_seconds(self._data["timecode"], self.year)

    @cached_property
    def _chunks(self):
        """Get the best chunks for this data."""
        return get_avhrr_lac_chunks((self._data.shape[0], 2048), float)

    @cached_property
    def _data(self):
        """Get the data."""
        return self.read()

    def read(self):
        """Read the file."""
        with open(self.filename, "rb") as fp_:
            data = np.memmap(fp_, dtype=scanline_dtype, mode="r")
        if np.all(np.median(data["frame_sync"], axis=0) > 1024):
            data = data.view(data.dtype.newbyteorder())
        return data

    @cached_property
    def platform_name(self):
        """Get the platform name."""
        return spacecrafts[np.median((self._data["id"]["id"] >> 3) & 15)]

    def get_dataset(self, key, info):
        """Get the dataset."""
        attrs = info.copy()
        attrs["platform_name"] = self.platform_name

        if key["name"] in ["latitude", "longitude"]:
            data = self._get_navigation_data(key)
        else:
            data = self._get_channel_data(key)

        result = xr.DataArray(data, dims=["y", "x"], attrs=attrs)
        mask = self._get_ch3_mask_or_true(key)
        return result.where(mask)

    def _get_channel_data(self, key):
        """Get channel data."""
        data = da.from_array(self._data["image_data"][:, :, _get_channel_index(key)], chunks=self._chunks)
        if key["calibration"] != "counts":
            if key["name"] in ["1", "2", "3a"]:
                data = self.calibrate_solar_channel(data, key)

            if key["name"] in ["3b", "4", "5"]:
                data = self.calibrate_thermal_channel(data, key)
        return data

    def _get_navigation_data(self, key):
        """Get navigation data."""
        lons, lats = self.lons_lats
        if key["name"] == "latitude":
            data = da.from_array(lats, chunks=self._chunks)
        else:
            data = da.from_array(lons, chunks=self._chunks)
        return data

    def _get_ch3_mask_or_true(self, key):
        mask = True
        if key["name"] == "3a":
            mask = np.tile(np.logical_not(self._is3b), (2048, 1)).T
        elif key["name"] == "3b":
            mask = np.tile(self._is3b, (2048, 1)).T
        return mask

    @cached_property
    def _is3b(self):
        return bfield(self._data["id"]["id"], 10) == 0

    def calibrate_thermal_channel(self, data, key):
        """Calibrate a thermal channel."""
        from pygac.calibration import calibrate_thermal
        line_numbers = (
            np.round((self.times - self.times[-1]) /
                     np.timedelta64(166666667, "ns"))).astype(int)
        line_numbers -= line_numbers[0]
        prt, ict, space = self.telemetry
        index = _get_channel_index(key)
        data = calibrate_thermal(data, prt, ict[:, index - 2],
                                 space[:, index], line_numbers,
                                 index + 1, self.calibrator)
        return data

    def calibrate_solar_channel(self, data, key):
        """Calibrate a solar channel."""
        from pygac.calibration import calibrate_solar
        julian_days = ((np.datetime64(self.start_time)
                        - np.datetime64(str(self.year) + "-01-01T00:00:00"))
                       / np.timedelta64(1, "D"))
        data = calibrate_solar(data, _get_channel_index(key), self.year, julian_days,
                               self.calibrator)
        return data

    @cached_property
    def calibrator(self):
        """Create a calibrator for the data."""
        from pygac.calibration import Calibrator
        pg_spacecraft = "".join(self.platform_name.split()).lower()
        return Calibrator(pg_spacecraft)

    @cached_property
    def telemetry(self):
        """Get the telemetry."""
        # This isn't converted to dask arrays as it does not work with pygac
        prt = np.mean(self._data["telemetry"]["PRT"], axis=1)
        ict = np.mean(self._data["back_scan"], axis=1)
        space = np.mean(self._data["space_data"][:, :], axis=1)

        return prt, ict, space

    @cached_property
    def lons_lats(self):
        """Get the lons and lats."""
        scanline_nb = len(self.times)
        scan_points = np.arange(0, 2048, 32)
        lons, lats = self._get_avhrr_tiepoints(scan_points, scanline_nb)

        lons, lats = geo_interpolate(
            lons.reshape((scanline_nb, -1)), lats.reshape((scanline_nb, -1)))
        return lons, lats

    def _get_avhrr_tiepoints(self, scan_points, scanline_nb):
        sgeom = avhrr(scanline_nb, scan_points, apply_offset=False)
        # no attitude error
        rpy = [0, 0, 0]
        s_times = sgeom.times(self.times[:, np.newaxis])
        orb = Orbital(self.platform_name)
        pixels_pos = compute_pixels(orb, sgeom, s_times, rpy)
        lons, lats, alts = get_lonlatalt(pixels_pos, s_times)
        return lons, lats

    @property
    def start_time(self):
        """Get the start time."""
        return time_seconds(self._data["timecode"][0, np.newaxis, :],
                            self.year).astype(dt.datetime)[0]

    @property
    def end_time(self):
        """Get the end time."""
        return time_seconds(self._data["timecode"][-1, np.newaxis, :],
                            self.year).astype(dt.datetime)[0]



# Copyright (c) 2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Tests for the writer config module."""
from __future__ import annotations

import os


class TestYAMLFiles:
    """Test and analyze the writer configuration files."""

    def test_filename_matches_writer_name(self):
        """Test that every writer filename matches the name in the YAML."""
        import yaml

        class IgnoreLoader(yaml.SafeLoader):

            def _ignore_all_tags(self, tag_suffix, node):
                return tag_suffix + " " + node.value
        IgnoreLoader.add_multi_constructor("", IgnoreLoader._ignore_all_tags)

        from satpy._config import glob_config
        from satpy.writers.core.config import read_writer_config
        for writer_config in glob_config("writers/*.yaml"):
            writer_fn = os.path.basename(writer_config)
            writer_fn_name = os.path.splitext(writer_fn)[0]
            writer_info = read_writer_config([writer_config],
                                             loader=IgnoreLoader)
            assert writer_fn_name == writer_info["name"]

    def test_available_writers(self):
        """Test the 'available_writers' function."""
        from satpy import available_writers
        writer_names = available_writers()
        assert len(writer_names) > 0
        assert isinstance(writer_names[0], str)
        assert "geotiff" in writer_names

        writer_infos = available_writers(as_dict=True)
        assert len(writer_names) == len(writer_infos)
        assert isinstance(writer_infos[0], dict)
        for writer_info in writer_infos:
            assert "name" in writer_info



# Copyright (c) 2017-2025 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Context managers for enhancements."""

import logging
from functools import wraps

import dask.array as da
import numpy as np
import xarray as xr

LOG = logging.getLogger(__name__)

def exclude_alpha(func):
    """Exclude the alpha channel from the DataArray before further processing."""

    @wraps(func)
    def wrapper(data, **kwargs):
        bands = data.coords["bands"].values
        exclude = ["A"] if "A" in bands else []
        band_data = data.sel(bands=[b for b in bands
                                    if b not in exclude])
        band_data = func(band_data, **kwargs)

        attrs = data.attrs
        attrs.update(band_data.attrs)
        # combine the new data with the excluded data
        new_data = xr.concat([band_data, data.sel(bands=exclude)],
                             dim="bands")
        data.data = new_data.sel(bands=bands).data
        data.attrs = attrs
        return data

    return wrapper


def on_separate_bands(func):
    """Apply `func` one band of the DataArray at a time.

    If this decorator is to be applied along with `on_dask_array`, this decorator has to be applied first, eg::

        @on_separate_bands
        @on_dask_array
        def my_enhancement_function(data):
            ...


    """

    @wraps(func)
    def wrapper(data, **kwargs):
        attrs = data.attrs
        data_arrs = []
        for idx, band in enumerate(data.coords["bands"].values):
            band_data = func(data.sel(bands=[band]), index=idx, **kwargs)
            data_arrs.append(band_data)
            # we assume that the func can add attrs
            attrs.update(band_data.attrs)
        data.data = xr.concat(data_arrs, dim="bands").data
        data.attrs = attrs
        return data

    return wrapper


def on_dask_array(func):
    """Pass the underlying dask array to *func* instead of the xarray.DataArray."""

    @wraps(func)
    def wrapper(data, **kwargs):
        dims = data.dims
        coords = data.coords
        d_arr = func(data.data, **kwargs)
        return xr.DataArray(d_arr, dims=dims, coords=coords)

    return wrapper


def using_map_blocks(func):
    """Run the provided function using :func:`dask.array.map_blocks`.

    This means dask will call the provided function with a single chunk
    as a numpy array.
    """

    @wraps(func)
    def wrapper(data, **kwargs):
        return da.map_blocks(func, data, meta=np.array((), dtype=data.dtype), dtype=data.dtype, chunks=data.chunks,
                             **kwargs)

    return on_dask_array(wrapper)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for the Arctica-M1 MSU-GS/A data.

The files for this reader are HDF5 and contain channel data at 1km resolution
for the VIS channels and 4km resolution for the IR channels. Geolocation data
is available at both resolutions, as is sun and satellite geometry.

This reader was tested on sample data provided by EUMETSAT.

"""

import datetime as dt

import numpy as np

from satpy.readers.core.hdf5 import HDF5FileHandler


class MSUGSAFileHandler(HDF5FileHandler):
    """MSU-GS/A L1B file reader."""

    @property
    def start_time(self):
        """Time for timeslot scan start."""
        dtstr = self["/attr/timestamp_without_timezone"]
        return dt.datetime.strptime(dtstr, "%Y-%m-%dT%H:%M:%S")

    @property
    def satellite_altitude(self):
        """Satellite altitude at time of scan.

        There is no documentation but this appears to be
        height above surface in meters.
        """
        return float(self["/attr/satellite_observation_point_height"])

    @property
    def satellite_latitude(self):
        """Satellite latitude at time of scan."""
        return float(self["/attr/satellite_observation_point_latitude"])

    @property
    def satellite_longitude(self):
        """Satellite longitude at time of scan."""
        return float(self["/attr/satellite_observation_point_longitude"])

    @property
    def sensor_name(self):
        """Sensor name is hardcoded."""
        sensor = "msu_gsa"
        return sensor

    @property
    def platform_name(self):
        """Platform name is also hardcoded."""
        platform = "Arctica-M-N1"
        return platform

    @staticmethod
    def _apply_scale_offset(in_data):
        """Apply the scale and offset to data."""
        scl = in_data.attrs["scale"]
        off = in_data.attrs["offset"]
        return in_data * scl + off

    def get_dataset(self, dataset_id, ds_info):
        """Load data variable and metadata and calibrate if needed."""
        file_key = ds_info.get("file_key", dataset_id["name"])
        data = self[file_key]
        attrs = data.attrs.copy()  # avoid contaminating other band loading
        attrs.update(ds_info)

        # The fill value also needs to be applied
        fill_val = attrs.pop("fill_value")
        data = data.where(data != fill_val, np.nan)

        # Data has a scale and offset that we must apply
        data = self._apply_scale_offset(data)

        # Data is given as radiance values, we must convert if we want reflectance
        if dataset_id.get("calibration") == "reflectance":
            solconst = float(attrs.pop("F_solar_constant"))
            data = np.pi * data / solconst
            # Satpy expects reflectance values in 0-100 range
            data = data * 100.

        data.attrs = attrs
        data.attrs.update({
            "platform_name": self.platform_name,
            "sensor": self.sensor_name,
            "sat_altitude": self.satellite_altitude,
            "sat_latitude": self.satellite_latitude,
            "sat_longitude": self.satellite_longitude,
        })

        return data



"""Reader for GMS-5 VISSR Level 1B data.

Introduction
------------
The ``gms5_vissr_l1b`` reader can decode, navigate and calibrate Level 1B data
from the Visible and Infrared Spin Scan Radiometer (VISSR) in `VISSR
archive format`. Corresponding platforms are GMS-5 (Japanese Geostationary
Meteorological Satellite) and GOES-09 (2003-2006 backup after MTSAT-1 launch
failure).

VISSR has four channels, each stored in a separate file:

.. code-block:: none

    VISSR_20020101_0031_IR1.A.IMG
    VISSR_20020101_0031_IR2.A.IMG
    VISSR_20020101_0031_IR3.A.IMG
    VISSR_20020101_0031_VIS.A.IMG

This is how to read them with Satpy:

.. code-block:: python

    from satpy import Scene
    import glob

    filenames = glob.glob(""/data/VISSR*")
    scene = Scene(filenames, reader="gms5-vissr_l1b")
    scene.load(["VIS", "IR1"])


References:
~~~~~~~~~~~

Details about platform, instrument and data format can be found in the
following references:

    - `VISSR Format Description`_
    - `GMS User Guide`_

.. _VISSR Format Description:
    https://www.data.jma.go.jp/mscweb/en/operation/fig/VISSR_FORMAT_GMS-5.pdf
.. _GMS User Guide:
    https://www.data.jma.go.jp/mscweb/en/operation/fig/GMS_Users_Guide_3rd_Edition_Rev1.pdf


Compression
-----------

Gzip-compressed VISSR files can be decompressed on the fly using
:class:`~satpy.readers.core.remote.FSFile`:

.. code-block:: python

    import fsspec
    from satpy import Scene
    from satpy.readers.core.remote import FSFile

    filename = "VISSR_19960217_2331_IR1.A.IMG.gz"
    open_file = fsspec.open(filename, compression="gzip")
    fs_file = FSFile(open_file)
    scene = Scene([fs_file], reader="gms5-vissr_l1b")
    scene.load(["IR1"])


Calibration
-----------

Sensor counts are calibrated by looking up reflectance/temperature values in the
calibration tables included in each file. See section 2.2 in the VISSR user
guide.


Navigation
----------

VISSR images are oversampled and not rectified.


Oversampling
~~~~~~~~~~~~
VISSR oversamples the viewed scene in E-W direction by a factor of ~1.46:
IR/VIS pixels are 14/3.5 urad on a side, but the instrument samples every
9.57/2.39 urad in E-W direction. That means pixels are actually overlapping on
the ground.

This cannot be represented by a pyresample area definition, so each dataset
is accompanied by 2-dimensional longitude and latitude coordinates. For
resampling purpose a full disc area definition with uniform sampling is provided
via

.. code-block:: python

    scene[dataset].attrs["area_def_uniform_sampling"]


Rectification
~~~~~~~~~~~~~

VISSR images are not rectified. That means lon/lat coordinates are different

1) for all channels of the same repeat cycle, even if their spatial resolution
   is identical (IR channels)
2) for different repeat cycles, even if the channel is identical

However, the above area definition is using the nominal subsatellite point as
projection center. As this rarely changes, the area definition is pretty
constant.


Performance
~~~~~~~~~~~

Navigation of VISSR images is computationally expensive, because for each pixel
the view vector of the (rotating) instrument needs to be intersected with the
earth, including interpolation of attitude and orbit prediction. For IR channels
this takes about 10 seconds, for VIS channels about 160 seconds.


Space Pixels
------------

VISSR produces data for pixels outside the Earth disk (i.e. atmospheric limb or
deep space pixels). By default, these pixels are masked out as they contain
data of limited or no value, but some applications do require these pixels.
To turn off masking, set ``mask_space=False`` upon scene creation:

.. code-block:: python

    import satpy
    import glob

    filenames = glob.glob("VISSR*.IMG")
    scene = satpy.Scene(filenames,
                        reader="gms5-vissr_l1b",
                        reader_kwargs={"mask_space": False})
    scene.load(["VIS", "IR1])


Metadata
--------

Dataset attributes include metadata such as time and orbital parameters,
see :ref:`dataset_metadata`.

Partial Scans
-------------

Between 2001 and 2003 VISSR also recorded partial scans of the northern
hemisphere. On demand a special Typhoon schedule would be activated between
03:00 and 05:00 UTC.
"""

import datetime as dt

import dask.array as da
import numba
import numpy as np
import xarray as xr

import satpy.readers.core._geos_area as geos_area
import satpy.readers.gms.gms5_vissr_format as fmt
import satpy.readers.gms.gms5_vissr_navigation as nav
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.utils import generic_open
from satpy.readers.hrit_jma import mjd2datetime64
from satpy.utils import datetime64_to_pydatetime, get_legacy_chunk_size

CHUNK_SIZE = get_legacy_chunk_size()


def _recarr2dict(arr, preserve=None):
    if not preserve:
        preserve = []
    res = {}
    for key, value in zip(arr.dtype.names, arr):
        if key.startswith("reserved"):
            continue
        if value.dtype.names and key not in preserve:
            # Nested record array
            res[key] = _recarr2dict(value)
        else:
            # Scalar or record array that shall be preserved
            res[key] = value
    return res


class GMS5VISSRFileHandler(BaseFileHandler):
    """File handler for GMS-5 VISSR data in VISSR archive format."""

    def __init__(self, filename, filename_info, filetype_info, mask_space=True):
        """Initialize the file handler.

        Args:
            filename: Name of file to be read
            filename_info: Information obtained from filename
            filetype_info: Information about file type
            mask_space: Mask space pixels.
        """
        super(GMS5VISSRFileHandler, self).__init__(
            filename, filename_info, filetype_info
        )
        self._filename = filename
        self._filename_info = filename_info
        self._header, self._channel_type = self._read_header(filename)
        self._mda = self._get_mda()
        self._mask_space = mask_space

    def _read_header(self, filename):
        header = {}
        with generic_open(filename, mode="rb") as file_obj:
            header["control_block"] = self._read_control_block(file_obj)
            channel_type = self._get_channel_type(
                header["control_block"]["parameter_block_size"]
            )
            header["image_parameters"] = self._read_image_params(file_obj, channel_type)
        return header, channel_type

    @staticmethod
    def _get_channel_type(parameter_block_size):
        if parameter_block_size == 4:
            return fmt.VIS_CHANNEL
        elif parameter_block_size == 16:
            return fmt.IR_CHANNEL
        raise ValueError(
            f"Cannot determine channel type, possibly corrupt file "
            f"(unknown parameter block size: {parameter_block_size})"
        )

    def _read_control_block(self, file_obj):
        ctrl_block = read_from_file_obj(file_obj, dtype=fmt.CONTROL_BLOCK, count=1)
        return _recarr2dict(ctrl_block[0])

    def _read_image_params(self, file_obj, channel_type):
        """Read image parameters from the header."""
        image_params = {}
        for name, param in fmt.IMAGE_PARAMS.items():
            image_params[name] = self._read_image_param(file_obj, param, channel_type)

        image_params["orbit_prediction"] = self._concat_orbit_prediction(
            image_params.pop("orbit_prediction_1"),
            image_params.pop("orbit_prediction_2"),
        )
        return image_params

    @staticmethod
    def _read_image_param(file_obj, param, channel_type):
        """Read a single image parameter block from the header."""
        image_params = read_from_file_obj(
            file_obj,
            dtype=param["dtype"],
            count=1,
            offset=param["offset"][channel_type],
        )
        return _recarr2dict(image_params[0], preserve=param.get("preserve"))

    @staticmethod
    def _concat_orbit_prediction(orb_pred_1, orb_pred_2):
        """Concatenate orbit prediction data.

        It is split over two image parameter blocks in the header.
        """
        orb_pred = orb_pred_1
        orb_pred["data"] = np.concatenate([orb_pred_1["data"], orb_pred_2["data"]])
        return orb_pred

    def _get_frame_parameters_key(self):
        if self._channel_type == fmt.VIS_CHANNEL:
            return "vis_frame_parameters"
        return "ir_frame_parameters"

    def _get_actual_shape(self):
        actual_num_lines = self._header["control_block"][
            "available_block_size_of_image_data"
        ]
        _, nominal_num_pixels = self._get_nominal_shape()
        return actual_num_lines, nominal_num_pixels

    def _get_nominal_shape(self):
        frame_params = self._header["image_parameters"]["mode"][
            self._get_frame_parameters_key()
        ]
        return frame_params["number_of_lines"], frame_params["number_of_pixels"]

    def _get_mda(self):
        return {
            "platform": self._mode_block["satellite_name"].decode().strip().upper(),
            "sensor": "VISSR",
            "time_parameters": self._get_time_parameters(),
            "orbital_parameters": self._get_orbital_parameters(),
        }

    def _get_orbital_parameters(self):
        # Note: SSP longitude in simple coordinate conversion table seems to be
        # incorrect (80 deg instead of 140 deg). Use orbital parameters instead.
        im_params = self._header["image_parameters"]
        mode = im_params["mode"]
        simple_coord = im_params["simple_coordinate_conversion_table"]
        orb_params = im_params["coordinate_conversion"]["orbital_parameters"]
        return {
            "satellite_nominal_longitude": mode["ssp_longitude"],
            "satellite_nominal_latitude": 0.0,
            "satellite_nominal_altitude": mode["satellite_height"],
            "satellite_actual_longitude": orb_params["longitude_of_ssp"],
            "satellite_actual_latitude": orb_params["latitude_of_ssp"],
            "satellite_actual_altitude": simple_coord["satellite_height"],
        }

    def _get_time_parameters(self):
        start_time = self._get_start_time()
        end_time = start_time + dt.timedelta(
            minutes=25
        )  # Source: GMS User Guide, section 3.3.1
        return {
            "nominal_start_time": start_time,
            "nominal_end_time": end_time,
        }

    def _get_start_time(self):
        start_time = mjd2datetime64(self._mode_block["observation_time_mjd"])
        start_time = datetime64_to_pydatetime(start_time)
        return start_time.replace(second=0, microsecond=0)

    def get_dataset(self, dataset_id, ds_info):
        """Get dataset from file."""
        image_data = self._get_image_data()
        counts = self._get_counts(image_data)
        dataset = self._calibrate(counts, dataset_id)
        space_masker = SpaceMasker(image_data, dataset_id["name"])
        dataset = self._mask_space_pixels(dataset, space_masker)
        self._attach_lons_lats(dataset, dataset_id)
        self._update_attrs(dataset, dataset_id, ds_info)
        return dataset

    def _get_image_data(self):
        image_data = self._read_image_data()
        return da.from_array(image_data, chunks=(CHUNK_SIZE,))

    def _read_image_data(self):
        num_lines, _ = self._get_actual_shape()
        specs = self._get_image_data_type_specs()
        with generic_open(self._filename, "rb") as file_obj:
            return read_from_file_obj(
                file_obj, dtype=specs["dtype"], count=num_lines, offset=specs["offset"]
            )

    def _get_image_data_type_specs(self):
        return fmt.IMAGE_DATA[self._channel_type]

    def _get_counts(self, image_data):
        return self._make_counts_data_array(image_data)

    def _make_counts_data_array(self, image_data):
        return xr.DataArray(
            image_data["image_data"],
            dims=("y", "x"),
            coords={
                "acq_time": ("y", self._get_acq_time(image_data)),
                "line_number": ("y", self._get_line_number(image_data)),
            },
        )

    def _get_acq_time(self, dask_array):
        acq_time = dask_array["LCW"]["scan_time"].compute()
        return mjd2datetime64(acq_time)

    def _get_line_number(self, dask_array):
        return dask_array["LCW"]["line_number"].compute()

    def _calibrate(self, counts, dataset_id):
        table = self._get_calibration_table(dataset_id)
        cal = Calibrator(table)
        return cal.calibrate(counts, dataset_id["calibration"])

    def _get_calibration_table(self, dataset_id):
        tables = {
            "VIS": self._header["image_parameters"]["vis_calibration"][
                "vis1_calibration_table"
            ]["brightness_albedo_conversion_table"],
            "IR1": self._header["image_parameters"]["ir1_calibration"][
                "conversion_table_of_equivalent_black_body_temperature"
            ],
            "IR2": self._header["image_parameters"]["ir2_calibration"][
                "conversion_table_of_equivalent_black_body_temperature"
            ],
            "IR3": self._header["image_parameters"]["wv_calibration"][
                "conversion_table_of_equivalent_black_body_temperature"
            ],
        }
        return tables[dataset_id["name"]]

    def _get_area_def_uniform_sampling(self, dataset_id):
        a = AreaDefEstimator(
            coord_conv_params=self._header["image_parameters"]["coordinate_conversion"],
            metadata=self._mda,
        )
        return a.get_area_def_uniform_sampling(dataset_id)

    def _mask_space_pixels(self, dataset, space_masker):
        if self._mask_space:
            return space_masker.mask_space(dataset)
        return dataset

    def _attach_lons_lats(self, dataset, dataset_id):
        lons, lats = self._get_lons_lats(dataset, dataset_id)
        dataset.coords["lon"] = lons
        dataset.coords["lat"] = lats

    def _get_lons_lats(self, dataset, dataset_id):
        lines, pixels = self._get_image_coords(dataset)
        nav_params = self._get_navigation_parameters(dataset_id)
        lons, lats = nav.get_lons_lats(lines, pixels, nav_params)
        return self._make_lons_lats_data_array(lons, lats)

    def _get_image_coords(self, data):
        lines = data.coords["line_number"].values
        pixels = np.arange(data.shape[1])
        return lines.astype(np.float64), pixels.astype(np.float64)

    def _get_navigation_parameters(self, dataset_id):
        return nav.ImageNavigationParameters(
            static=self._get_static_navigation_params(dataset_id),
            predicted=self._get_predicted_navigation_params()
        )

    def _get_static_navigation_params(self, dataset_id):
        """Get static navigation parameters.

        Note that, "central_line_number_of_vissr_frame" is different for each
        channel, even if their spatial resolution is identical. For example:

        VIS: 5513.0
        IR1: 1378.5
        IR2: 1378.7
        IR3: 1379.1001
        """
        alt_ch_name = _get_alternative_channel_name(dataset_id)
        scan_params = nav.ScanningParameters(
            start_time_of_scan=self._coord_conv["scheduled_observation_time"],
            spinning_rate=self._mode_block["spin_rate"],
            num_sensors=self._coord_conv["number_of_sensor_elements"][alt_ch_name],
            sampling_angle=self._coord_conv["sampling_angle_along_pixel"][alt_ch_name],
        )
        proj_params = self._get_proj_params(dataset_id)
        return nav.StaticNavigationParameters(
            proj_params=proj_params,
            scan_params=scan_params
        )

    def _get_proj_params(self, dataset_id):
        proj_params = nav.ProjectionParameters(
            image_offset=self._get_image_offset(dataset_id),
            scanning_angles=self._get_scanning_angles(dataset_id),
            earth_ellipsoid=self._get_earth_ellipsoid()
        )
        return proj_params

    def _get_earth_ellipsoid(self):
        # Use earth radius and flattening from JMA's Msial library, because
        # the values in the data seem to be pretty old. For example the
        # equatorial radius is from the Bessel Ellipsoid (1841).
        return nav.EarthEllipsoid(
            flattening=nav.EARTH_FLATTENING,
            equatorial_radius=nav.EARTH_EQUATORIAL_RADIUS,
        )

    def _get_scanning_angles(self, dataset_id):
        alt_ch_name = _get_alternative_channel_name(dataset_id)
        misalignment = np.ascontiguousarray(
            self._coord_conv["matrix_of_misalignment"].transpose().astype(np.float64)
        )
        return nav.ScanningAngles(
            stepping_angle=self._coord_conv["stepping_angle_along_line"][alt_ch_name],
            sampling_angle=self._coord_conv["sampling_angle_along_pixel"][
                alt_ch_name],
            misalignment=misalignment
        )

    def _get_image_offset(self, dataset_id):
        alt_ch_name = _get_alternative_channel_name(dataset_id)
        center_line_vissr_frame = self._coord_conv["central_line_number_of_vissr_frame"][
            alt_ch_name
        ]
        center_pixel_vissr_frame = self._coord_conv["central_pixel_number_of_vissr_frame"][
            alt_ch_name
        ]
        pixel_offset = self._coord_conv[
            "pixel_difference_of_vissr_center_from_normal_position"
        ][alt_ch_name]
        return nav.ImageOffset(
            line_offset=center_line_vissr_frame,
            pixel_offset=center_pixel_vissr_frame + pixel_offset,
        )

    def _get_predicted_navigation_params(self):
        """Get predictions of time-dependent navigation parameters."""
        attitude_prediction = self._get_attitude_prediction()
        orbit_prediction = self._get_orbit_prediction()
        return nav.PredictedNavigationParameters(
            attitude=attitude_prediction,
            orbit=orbit_prediction
        )

    def _get_attitude_prediction(self):
        att_pred = self._header["image_parameters"]["attitude_prediction"]["data"]
        attitudes = nav.Attitude(
            angle_between_earth_and_sun=att_pred["sun_earth_angle"].astype(
                np.float64),
            angle_between_sat_spin_and_z_axis=att_pred[
                "right_ascension_of_attitude"
            ].astype(np.float64),
            angle_between_sat_spin_and_yz_plane=att_pred[
                "declination_of_attitude"
            ].astype(np.float64),
        )
        attitude_prediction = nav.AttitudePrediction(
            prediction_times=att_pred["prediction_time_mjd"].astype(np.float64),
            attitude=attitudes
        )
        return attitude_prediction

    def _get_orbit_prediction(self):
        orb_pred = self._header["image_parameters"]["orbit_prediction"]["data"]
        orbit_angles = nav.OrbitAngles(
            greenwich_sidereal_time=np.deg2rad(
                orb_pred["greenwich_sidereal_time"].astype(np.float64)
            ),
            declination_from_sat_to_sun=np.deg2rad(
                orb_pred["sat_sun_vector_earth_fixed"]["elevation"].astype(np.float64)
            ),
            right_ascension_from_sat_to_sun=np.deg2rad(
                orb_pred["sat_sun_vector_earth_fixed"]["azimuth"].astype(np.float64)
            ),
        )
        sat_position = nav.Satpos(
            x=orb_pred["satellite_position_earth_fixed"][:, 0].astype(np.float64),
            y=orb_pred["satellite_position_earth_fixed"][:, 1].astype(np.float64),
            z=orb_pred["satellite_position_earth_fixed"][:, 2].astype(np.float64),
        )
        orbit_prediction = nav.OrbitPrediction(
            prediction_times=orb_pred["prediction_time_mjd"].astype(np.float64),
            angles=orbit_angles,
            sat_position=sat_position,
            nutation_precession=np.ascontiguousarray(
                orb_pred["conversion_matrix"].transpose(0, 2, 1).astype(np.float64)
            ),
        )
        return orbit_prediction

    def _make_lons_lats_data_array(self, lons, lats):
        lons = xr.DataArray(
            lons,
            dims=("y", "x"),
            attrs={"standard_name": "longitude", "units": "degrees_east"},
        )
        lats = xr.DataArray(
            lats,
            dims=("y", "x"),
            attrs={"standard_name": "latitude", "units": "degrees_north"},
        )
        return lons, lats

    def _update_attrs(self, dataset, dataset_id, ds_info):
        dataset.attrs.update(ds_info)
        dataset.attrs.update(self._mda)
        dataset.attrs[
            "area_def_uniform_sampling"
        ] = self._get_area_def_uniform_sampling(dataset_id)

    @property
    def start_time(self):
        """Nominal start time of the dataset."""
        return self._mda["time_parameters"]["nominal_start_time"]

    @property
    def end_time(self):
        """Nominal end time of the dataset."""
        return self._mda["time_parameters"]["nominal_end_time"]

    @property
    def _coord_conv(self):
        return self._header["image_parameters"]["coordinate_conversion"]

    @property
    def _mode_block(self):
        return self._header["image_parameters"]["mode"]


def _get_alternative_channel_name(dataset_id):
    return fmt.ALT_CHANNEL_NAMES[dataset_id["name"]]


def read_from_file_obj(file_obj, dtype, count, offset=0):
    """Read data from file object.

    Args:
        file_obj: An open file object.
        dtype: Data type to be read.
        count: Number of elements to be read.
        offset: Byte offset where to start reading.
    """
    file_obj.seek(offset)
    data = file_obj.read(dtype.itemsize * count)
    return np.frombuffer(data, dtype=dtype, count=count)


class Calibrator:
    """Calibrate VISSR data to reflectance or brightness temperature.

    Reference: Section 2.2 in the VISSR User Guide.
    """

    def __init__(self, calib_table):
        """Initialize the calibrator.

        Args:
            calib_table: Calibration table
        """
        self._calib_table = calib_table

    def calibrate(self, counts, calibration):
        """Transform counts to given calibration level."""
        if calibration == "counts":
            return counts
        res = self._calibrate(counts)
        res = self._postproc(res, calibration)
        return self._make_data_array(res, counts)

    def _calibrate(self, counts):
        return da.map_blocks(
            self._lookup_calib_table,
            counts.data,
            calib_table=self._calib_table,
            dtype=np.float32,
        )

    def _postproc(self, res, calibration):
        if calibration == "reflectance":
            res = self._convert_to_percent(res)
        return res

    def _convert_to_percent(self, res):
        return res * 100

    def _make_data_array(self, interp, counts):
        return xr.DataArray(
            interp,
            dims=counts.dims,
            coords=counts.coords,
        )

    def _lookup_calib_table(self, counts, calib_table):
        return calib_table[counts]


class SpaceMasker:
    """Mask pixels outside the earth disk."""

    _fill_value = -1  # scanline not intersecting the earth

    def __init__(self, image_data, channel):
        """Initialize the space masker.

        Args:
            image_data: Image data
            channel: Channel name
        """
        self._image_data = image_data
        self._channel = channel
        self._shape = image_data["image_data"].shape
        self._earth_mask = self._get_earth_mask()

    def mask_space(self, dataset):
        """Mask space pixels in the given dataset."""
        return dataset.where(self._earth_mask).astype(np.float32)

    def _get_earth_mask(self):
        earth_edges = self._get_earth_edges()
        return get_earth_mask(self._shape, earth_edges, self._fill_value)

    def _get_earth_edges(self):
        west_edges = self._get_earth_edges_per_scan_line("west_side_earth_edge")
        east_edges = self._get_earth_edges_per_scan_line("east_side_earth_edge")
        return west_edges, east_edges

    def _get_earth_edges_per_scan_line(self, cardinal):
        edges = self._image_data["LCW"][cardinal].compute().astype(np.int32)
        if is_vis_channel(self._channel):
            edges = self._correct_vis_edges(edges)
        return edges

    def _correct_vis_edges(self, edges):
        """Correct VIS edges.

        VIS data contains earth edges of IR channel. Compensate for that
        by scaling with a factor of 4 (1 IR pixel ~ 4 VIS pixels).
        """
        return np.where(edges != self._fill_value, edges * 4, edges)


@numba.njit
def get_earth_mask(shape, earth_edges, fill_value=-1):
    """Get binary mask where 1/0 indicates earth/space.

    Args:
        shape: Image shape
        earth_edges: First and last earth pixel in each scanline
        fill_value: Fill value for scanlines not intersecting the earth.
    """
    first_earth_pixels, last_earth_pixels = earth_edges
    mask = np.zeros(shape, dtype=np.int8)
    for line in range(shape[0]):
        first = first_earth_pixels[line]
        last = last_earth_pixels[line]
        if first == fill_value or last == fill_value:
            continue
        mask[line, first:last+1] = 1
    return mask


def is_vis_channel(channel_name):
    """Check if it's the visible channel."""
    return channel_name == "VIS"


class AreaDefEstimator:
    """Estimate area definition for VISSR images."""

    full_disk_size = {
        "IR": 2366,
        "VIS": 9464,
    }

    def __init__(self, coord_conv_params, metadata):
        """Initialize the area definition estimator.

        Args:
            coord_conv_params: Coordinate conversion parameters
            metadata: VISSR file metadata
        """
        self.coord_conv = coord_conv_params
        self.metadata = metadata

    def get_area_def_uniform_sampling(self, dataset_id):
        """Get full disk area definition with uniform sampling.

        Args:
            dataset_id: ID of the corresponding dataset.
        """
        proj_dict = self._get_proj_dict(dataset_id)
        extent = geos_area.get_area_extent(proj_dict)
        return geos_area.get_area_definition(proj_dict, extent)

    def _get_proj_dict(self, dataset_id):
        proj_dict = {}
        proj_dict.update(self._get_name_dict(dataset_id))
        proj_dict.update(self._get_proj4_dict())
        proj_dict.update(self._get_shape_dict(dataset_id))
        return proj_dict

    def _get_name_dict(self, dataset_id):
        name_dict = geos_area.get_geos_area_naming(
            {
                "platform_name": self.metadata["platform"],
                "instrument_name": self.metadata["sensor"],
                "service_name": "western-pacific",
                "service_desc": "Western Pacific",
                "resolution": dataset_id["resolution"],
            }
        )
        return {
            "a_name": name_dict["area_id"],
            "p_id": name_dict["area_id"],
            "a_desc": name_dict["description"],
        }

    def _get_proj4_dict(
        self,
    ):
        # Use nominal parameters to make the area def as constant as possible
        return {
            "ssp_lon": self.metadata["orbital_parameters"][
                "satellite_nominal_longitude"
            ],
            "a": nav.EARTH_EQUATORIAL_RADIUS,
            "b": nav.EARTH_POLAR_RADIUS,
            "h": self.metadata["orbital_parameters"]["satellite_nominal_altitude"],
        }

    def _get_shape_dict(self, dataset_id):
        # Apply sampling from the vertical dimension to the horizontal
        # dimension to obtain a square area definition with uniform sampling.
        ch_type = fmt.CHANNEL_TYPES[dataset_id["name"]]
        alt_ch_name = _get_alternative_channel_name(dataset_id)
        stepping_angle = self.coord_conv["stepping_angle_along_line"][alt_ch_name]
        size = self.full_disk_size[ch_type]
        line_pixel_offset = 0.5 * size
        lfac_cfac = geos_area.sampling_to_lfac_cfac(stepping_angle)
        return {
            "nlines": size,
            "ncols": size,
            "lfac": lfac_cfac,
            "cfac": lfac_cfac,
            "coff": line_pixel_offset,
            "loff": line_pixel_offset,
            "scandir": "N2S",
        }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""VIIRS Active Fires reader.

This module implements readers for VIIRS Active Fires NetCDF and
ASCII files.
"""

import dask.dataframe as dd
import xarray as xr

from satpy.dataset.dataid import DataID
from satpy.readers.core.file_handlers import BaseFileHandler
from satpy.readers.core.netcdf import NetCDF4FileHandler

# map platform attributes to Oscar standard name
PLATFORM_MAP = {
    "NPP": "Suomi-NPP",
    "J01": "NOAA-20",
    "J02": "NOAA-21"
}


class VIIRSActiveFiresFileHandler(NetCDF4FileHandler):
    """NetCDF4 reader for VIIRS Active Fires."""

    def __init__(self, filename, filename_info, filetype_info,
                 auto_maskandscale=False, xarray_kwargs=None):
        """Open and perform initial investigation of NetCDF file."""
        super(VIIRSActiveFiresFileHandler, self).__init__(
            filename, filename_info, filetype_info,
            auto_maskandscale=auto_maskandscale, xarray_kwargs=xarray_kwargs)
        self.prefix = filetype_info.get("variable_prefix")

    def get_dataset(self, dsid: DataID, dsinfo: dict) -> xr.DataArray:  # noqa: D417
        """Get requested data as DataArray.

        Args:
            dsid: DataID to load data for
            dsinfo: Dataset information

        Returns:
            DataArray of the loaded variable

        """
        key = dsinfo.get("file_key", dsid["name"]).format(variable_prefix=self.prefix)
        data = self[key]
        # rename "phoney dims"
        data = data.rename(dict(zip(data.dims, ["y", "x"])))

        # handle attributes from YAML
        for key in ("units", "standard_name", "flag_meanings", "flag_values", "_FillValue"):
            # we only want to add information that isn't present already
            if key in dsinfo and key not in data.attrs:
                data.attrs[key] = dsinfo[key]
        if isinstance(data.attrs.get("flag_meanings"), str):
            data.attrs["flag_meanings"] = data.attrs["flag_meanings"].split(" ")

        # use more common CF standard units
        if data.attrs.get("units") == "kelvins":
            data.attrs["units"] = "K"

        data.attrs["platform_name"] = PLATFORM_MAP.get(self.filename_info["satellite_name"].upper(), "unknown")
        data.attrs["sensor"] = self.sensor_name

        return data

    @property
    def start_time(self):
        """Get first date/time when observations were recorded."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get last date/time when observations were recorded."""
        return self.filename_info.get("end_time", self.start_time)

    @property
    def sensor_name(self):
        """Name of sensor for this file."""
        return self["/attr/instrument_name"].lower()

    @property
    def platform_name(self):
        """Name of platform/satellite for this file."""
        return self["/attr/satellite_name"]


class VIIRSActiveFiresTextFileHandler(BaseFileHandler):
    """ASCII reader for VIIRS Active Fires."""

    def __init__(self, filename, filename_info, filetype_info):
        """Make sure filepath is valid and then reads data into a Dask DataFrame.

        Args:
            filename: Filename
            filename_info: Filename information
            filetype_info: Filetype information

        """
        skip_rows = filetype_info.get("skip_rows", 15)
        columns = filetype_info["columns"]
        self.file_content = dd.read_csv(filename, skiprows=skip_rows, header=None, names=columns)
        super(VIIRSActiveFiresTextFileHandler, self).__init__(filename, filename_info, filetype_info)
        self.platform_name = PLATFORM_MAP.get(self.filename_info["satellite_name"].upper(), "unknown")

    def get_dataset(self, dsid, dsinfo):
        """Get requested data as DataArray."""
        ds = self[dsid["name"]].to_dask_array(lengths=True)
        data = xr.DataArray(ds, dims=("y",), attrs={"platform_name": self.platform_name, "sensor": "VIIRS"})
        for key in ("units", "standard_name", "flag_meanings", "flag_values", "_FillValue"):
            # we only want to add information that isn't present already
            if key in dsinfo and key not in data.attrs:
                data.attrs[key] = dsinfo[key]
        if isinstance(data.attrs.get("flag_meanings"), str):
            data.attrs["flag_meanings"] = data.attrs["flag_meanings"].split(" ")
        return data

    @property
    def start_time(self):
        """Get first date/time when observations were recorded."""
        return self.filename_info["start_time"]

    @property
    def end_time(self):
        """Get last date/time when observations were recorded."""
        return self.filename_info.get("end_time", self.start_time)

    def __getitem__(self, key):
        """Get file content for 'key'."""
        return self.file_content[key]

    def __contains__(self, item):
        """Check if variable is in current file."""
        return item in self.file_content



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.

r"""Interface to SEVIRI L1B data from ICARE (Lille).

Introduction
------------

The ``seviri_l1b_icare`` reader reads MSG-SEVIRI L1.5 image data in HDF format
that has been produced by the ICARE Data and Services Center
Data can be accessed via: http://www.icare.univ-lille1.fr

Each SEVIRI timeslot comes as 12 HDF files, one per band. Only those bands
that are of interest need to be passed to the reader. Others can be ignored.
Filenames follow the format: GEO_L1B-MSG1_YYYY-MM-DDTHH-MM-SS_G_CHANN_VX-XX.hdf
Where:
YYYY, MM, DD, HH, MM, SS specify the timeslot starting time.
CHANN is the channel (i.e: HRV, IR016, WV073, etc)
VX-XX is the processing version number

Example:
--------
Here is an example how to read the data in satpy:

.. code-block:: python

    from satpy import Scene
    import glob

    filenames = glob.glob('data/*2019-03-01T12-00-00*.hdf')
    scn = Scene(filenames=filenames, reader='seviri_l1b_icare')
    scn.load(['VIS006', 'IR_108'])
    print(scn['IR_108'])

Output:

.. code-block:: none

    <xarray.DataArray 'array-a1d52b7e19ec5a875e2f038df5b60d7e' (y: 3712, x: 3712)>
    dask.array<add, shape=(3712, 3712), dtype=float32, chunksize=(1024, 1024), chunktype=numpy.ndarray>
    Coordinates:
        crs      object +proj=geos +a=6378169.0 +b=6356583.8 +lon_0=0.0 +h=35785831.0 +units=m +type=crs
      * y        (y) float64 5.566e+06 5.563e+06 5.56e+06 ... -5.566e+06 -5.569e+06
      * x        (x) float64 -5.566e+06 -5.563e+06 -5.56e+06 ... 5.566e+06 5.569e+06
    Attributes:
        start_time:           2004-12-29 12:15:00
        end_time:             2004-12-29 12:27:44
        area:                 Area ID: geosmsg\nDescription: MSG/SEVIRI low resol...
        name:                 IR_108
        resolution:           3000.403165817
        calibration:          brightness_temperature
        polarization:         None
        level:                None
        modifiers:            ()
        ancillary_variables:  []

"""

import datetime as dt

import numpy as np

from satpy.readers.core._geos_area import get_area_definition, get_area_extent
from satpy.readers.core.hdf4 import HDF4FileHandler


class SEVIRI_ICARE(HDF4FileHandler):
    """SEVIRI L1B handler for HDF4 files."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init the file handler."""
        super(SEVIRI_ICARE, self).__init__(filename,
                                           filename_info,
                                           filetype_info)
        # These are VIS bands
        self.ref_bands = ["HRV", "VIS006", "VIS008", "IR_016"]
        # And these are IR bands
        self.bt_bands = ["IR_039", "IR_062", "IR_073",
                         "IR_087", "IR_097", "IR_108",
                         "IR_120", "IR_134",
                         "WV_062", "WV_073"]

    @property
    def sensor_name(self):
        """Get the sensor name."""
        # the sensor and platform names are stored together, eg: MSG1/SEVIRI
        attr = self["/attr/Sensors"]
        if isinstance(attr, np.ndarray):
            attr = str(attr.astype(str)).lower()
        else:
            attr = attr.lower()
        plat = attr[0:4]
        sens = attr[5:]
        # icare uses non-standard platform names
        if plat == "msg1":
            plat = "Meteosat-08"
        elif plat == "msg2":
            plat = "Meteosat-09"
        elif plat == "msg3":
            plat = "Meteosat-10"
        elif plat == "msg4":
            plat = "Meteosat-11"
        else:
            raise NameError("Unsupported satellite platform:"+plat)
        return [plat, sens]

    @property
    def satlon(self):
        """Get the satellite longitude."""
        attr = self["/attr/Sub_Satellite_Longitude"]
        if isinstance(attr, np.ndarray):
            attr = float(attr.astype(str))
        return attr

    @property
    def projlon(self):
        """Get the projection longitude."""
        attr = self["/attr/Projection_Longitude"]
        if isinstance(attr, np.ndarray):
            attr = float(attr.astype(str))
        return attr

    @property
    def projection(self):
        """Get the projection."""
        attr = self["/attr/Geographic_Projection"]
        if isinstance(attr, np.ndarray):
            attr = str(attr.astype(str))
        attr = attr.lower()
        if attr != "geos":
            raise NotImplementedError("Only the GEOS projection is supported.\
                                        This is:", attr)
        return attr

    @property
    def zone(self):
        """Get the zone."""
        attr = self["/attr/Zone"]
        if isinstance(attr, np.ndarray):
            attr = str(attr.astype(str)).lower()
        return attr

    @property
    def res(self):
        """Get the resolution."""
        attr = self["/attr/Nadir_Pixel_Size"]
        if isinstance(attr, np.ndarray):
            attr = str(attr.astype(str)).lower()
        return float(attr)

    @property
    def end_time(self):
        """Get the end time."""
        attr = self["/attr/End_Acquisition_Date"]
        if isinstance(attr, np.ndarray):
            attr = str(attr.astype(str))
        # In some versions milliseconds are present, sometimes not.
        try:
            endacq = dt.datetime.strptime(attr, "%Y-%m-%dT%H:%M:%SZ")
        except ValueError:
            endacq = dt.datetime.strptime(attr, "%Y-%m-%dT%H:%M:%S.%fZ")
        return endacq

    @property
    def start_time(self):
        """Get the start time."""
        attr = self["/attr/Beginning_Acquisition_Date"]
        if isinstance(attr, np.ndarray):
            attr = str(attr.astype(str))
        # In some versions milliseconds are present, sometimes not.
        try:
            stacq = dt.datetime.strptime(attr, "%Y-%m-%dT%H:%M:%SZ")
        except ValueError:
            stacq = dt.datetime.strptime(attr, "%Y-%m-%dT%H:%M:%S.%fZ")
        return stacq

    @property
    def alt(self):
        """Get the altitude."""
        attr = self["/attr/Altitude"]
        if isinstance(attr, np.ndarray):
            attr = attr.astype(str)
        attr = float(attr)
        # This is stored in km, convert to m
        attr = attr * 1000.
        return attr

    @property
    def geoloc(self):
        """Get the geolocation."""
        attr = self["/attr/Geolocation"]
        if isinstance(attr, np.ndarray):
            attr = attr.astype(str)
        cfac = float(attr[0])
        coff = float(attr[1])
        lfac = float(attr[2])
        loff = float(attr[3])
        return [cfac, lfac, coff, loff]

    def get_metadata(self, data, ds_info):
        """Get the metadata."""
        mda = {}
        mda.update(data.attrs)
        mda.update(ds_info)
        geoloc = self.geoloc
        mda.update({
                    "start_time": self.start_time,
                    "end_time": self.end_time,
                    "platform_name": self.sensor_name[0],
                    "sensor": self.sensor_name[1],
                    "zone": self.zone,
                    "projection_altitude": self.alt,
                    "cfac": geoloc[0],
                    "lfac": geoloc[1],
                    "coff": geoloc[2],
                    "loff": geoloc[3],
                    "resolution": self.res,
                    "satellite_actual_longitude": self.satlon,
                    "projection_longitude": self.projlon,
                    "projection_type": self.projection
        })

        return mda

    def _get_dsname(self, ds_id):
        """Return the correct dataset name based on requested band."""
        if ds_id["name"] in self.ref_bands:
            ds_get_name = "Normalized_Radiance"
        elif ds_id["name"] in self.bt_bands:
            ds_get_name = "Brightness_Temperature"
        else:
            raise NameError("Datset type "+ds_id["name"]+" is not supported.")
        return ds_get_name

    def get_dataset(self, ds_id, ds_info):
        """Get the dataset."""
        ds_get_name = self._get_dsname(ds_id)
        data = self[ds_get_name]
        data.attrs = self.get_metadata(data, ds_info)
        fill = data.attrs.pop("_FillValue")
        offset = data.attrs.get("add_offset")
        scale_factor = data.attrs.get("scale_factor")
        data = data.where(data != fill)
        data = data.astype(np.float32)
        if scale_factor is not None and offset is not None:
            data = data * scale_factor
            data = data + offset
            # Now we correct range from 0-1 to 0-100 for VIS:
            if ds_id["name"] in self.ref_bands:
                data = data * 100.
        return data

    def get_area_def(self, ds_id):
        """Get the area def."""
        ds_get_name = self._get_dsname(ds_id)
        ds_shape = self[ds_get_name + "/shape"]
        geoloc = self.geoloc

        pdict = {}
        pdict["cfac"] = np.int32(geoloc[0])
        pdict["lfac"] = np.int32(geoloc[1])
        pdict["coff"] = np.float32(geoloc[2])
        pdict["loff"] = -np.float32(geoloc[3])

        # Unfortunately this dataset does not store a, b or h.
        # We assume a and b here, and calculate h from altitude
        # a and b are from SEVIRI data HRIT header (201912101300)
        pdict["a"] = 6378169
        pdict["b"] = 6356583.8
        pdict["h"] = self.alt - pdict["a"]
        pdict["ssp_lon"] = self.projlon
        pdict["ncols"] = int(ds_shape[0])
        pdict["nlines"] = int(ds_shape[1])

        # Force scandir to SEVIRI default, not known from file
        pdict["scandir"] = "S2N"
        pdict["a_name"] = "geosmsg"
        if ds_id["name"] == "HRV":
            pdict["a_desc"] = "MSG/SEVIRI HRV channel area"
            pdict["p_id"] = "msg_hires"
        else:
            pdict["a_desc"] = "MSG/SEVIRI low resolution channel area"
            pdict["p_id"] = "msg_lowres"

        aex = get_area_extent(pdict)
        area = get_area_definition(pdict, aex)

        return area



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017-2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Utilities for EUMETSAT satellite data."""

import datetime as dt

import numpy as np

# 6 bytes, 8 bytes, 10 bytes
time_cds_short = [("Days", ">u2"), ("Milliseconds", ">u4")]
time_cds = time_cds_short + [("Microseconds", ">u2")]
time_cds_expanded = time_cds + [("Nanoseconds", ">u2")]
issue_revision = [("Issue", np.uint16), ("Revision", np.uint16)]


def timecds2datetime(tcds):
    """Convert time_cds-variables to datetime-object.

    Works both with a dictionary and a numpy record_array.
    """
    days = int(tcds["Days"].item())
    milliseconds = int(tcds["Milliseconds"].item())
    try:
        microseconds = int(tcds["Microseconds"].item())
    except (KeyError, ValueError):
        microseconds = 0
    try:
        microseconds += int(tcds["Nanoseconds"].item()) / 1000.
    except (KeyError, ValueError):
        pass

    reference = dt.datetime(1958, 1, 1)
    delta = dt.timedelta(days=days, milliseconds=milliseconds,
                         microseconds=microseconds)

    return reference + delta


def recarray2dict(arr):
    """Convert numpy record array to a dictionary."""
    res = {}
    tcds_types = [time_cds_short, time_cds, time_cds_expanded]

    for dtuple in arr.dtype.descr:
        key = dtuple[0]
        ntype = dtuple[1]
        data = arr[key]
        if ntype in tcds_types:
            if data.size > 1:
                res[key] = np.array([timecds2datetime(item)
                                     for item in data.ravel()]).reshape(data.shape)
            else:
                res[key] = timecds2datetime(data)
        elif isinstance(ntype, list):
            res[key] = recarray2dict(data)
        else:
            if data.size == 1:
                data = data[0]
                if ntype[:2] == "|S":
                    # Python2 and Python3 handle strings differently
                    try:
                        data = data.decode()
                    except ValueError:
                        data = None
                    else:
                        data = data.split(":")[0].strip()
                res[key] = data
            else:
                res[key] = data.squeeze()

    return res


def get_service_mode(instrument_name, ssp_lon):
    """Get information about service mode for a given instrument and subsatellite longitude."""
    service_modes = {"seviri": {"0.0": {"service_name": "fes", "service_desc": "Full Earth Scanning service"},
                                "9.5": {"service_name": "rss", "service_desc": "Rapid Scanning Service"},
                                "41.5": {"service_name": "iodc", "service_desc": "Indian Ocean Data Coverage service"},
                                "45.5": {"service_name": "iodc", "service_desc": "Indian Ocean Data Coverage service"}
                                },
                     "fci": {"0.0": {"service_name": "fdss", "service_desc": "Full Disk Scanning Service"},
                             "9.5": {"service_name": "rss", "service_desc": "Rapid Scanning Service"},
                             },
                     }
    unknown_modes = {"service_name": "unknown", "service_desc": "unknown"}

    return service_modes.get(instrument_name, unknown_modes).get("{:.1f}".format(ssp_lon), unknown_modes)



# Copyright (c) 2010-2023 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit tests for resampling and crop-related functionality in scene.py."""
from unittest import mock

import numpy as np
import pytest
import xarray as xr
from dask import array as da

from satpy import Scene
from satpy.dataset.dataid import default_id_keys_config
from satpy.tests.utils import make_cid, make_dataid

# NOTE:
# The following fixtures are not defined in this file, but are used and injected by Pytest:
# - include_test_etc


class TestSceneCrop:
    """Test creating new Scenes by cropping an existing Scene."""

    def test_crop(self):
        """Test the crop method."""
        from pyresample.geometry import AreaDefinition
        scene1 = Scene()
        area_extent = (-5570248.477339745, -5561247.267842293, 5567248.074173927,
                       5570248.477339745)
        proj_dict = {"a": 6378169.0, "b": 6356583.8, "h": 35785831.0,
                     "lon_0": 0.0, "proj": "geos", "units": "m"}
        x_size = 3712
        y_size = 3712
        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            proj_dict,
            x_size,
            y_size,
            area_extent,
        )
        area_def2 = AreaDefinition(
            "test2",
            "test2",
            "test2",
            proj_dict,
            x_size // 2,
            y_size // 2,
            area_extent,
            )
        scene1["1"] = xr.DataArray(np.zeros((y_size, x_size)))
        scene1["2"] = xr.DataArray(np.zeros((y_size, x_size)), dims=("y", "x"))
        scene1["3"] = xr.DataArray(np.zeros((y_size, x_size)), dims=("y", "x"),
                                   attrs={"area": area_def})
        scene1["4"] = xr.DataArray(np.zeros((y_size // 2, x_size // 2)), dims=("y", "x"),
                                   attrs={"area": area_def2})

        # by area
        crop_area = AreaDefinition(
            "test",
            "test",
            "test",
            proj_dict,
            x_size,
            y_size,
            (area_extent[0] + 10000., area_extent[1] + 500000.,
             area_extent[2] - 10000., area_extent[3] - 500000.)
        )
        new_scn1 = scene1.crop(crop_area)
        assert "1" in new_scn1
        assert "2" in new_scn1
        assert "3" in new_scn1
        assert new_scn1["1"].shape == (y_size, x_size)
        assert new_scn1["2"].shape == (y_size, x_size)
        assert new_scn1["3"].shape == (3380, 3708)
        assert new_scn1["4"].shape == (1690, 1854)

        # by lon/lat bbox
        new_scn1 = scene1.crop(ll_bbox=(-20., -5., 0, 0))
        assert "1" in new_scn1
        assert "2" in new_scn1
        assert "3" in new_scn1
        assert new_scn1["1"].shape == (y_size, x_size)
        assert new_scn1["2"].shape == (y_size, x_size)
        assert new_scn1["3"].shape == (184, 714)
        assert new_scn1["4"].shape == (92, 357)

        # by x/y bbox
        new_scn1 = scene1.crop(xy_bbox=(-200000., -100000., 0, 0))
        assert "1" in new_scn1
        assert "2" in new_scn1
        assert "3" in new_scn1
        assert new_scn1["1"].shape == (y_size, x_size)
        assert new_scn1["2"].shape == (y_size, x_size)
        assert new_scn1["3"].shape == (36, 70)
        assert new_scn1["4"].shape == (18, 35)

    def test_crop_epsg_crs(self):
        """Test the crop method when source area uses an EPSG code."""
        from pyresample.geometry import AreaDefinition

        scene1 = Scene()
        area_extent = (699960.0, 5390220.0, 809760.0, 5500020.0)
        x_size = 3712
        y_size = 3712
        area_def = AreaDefinition(
            "test", "test", "test",
            "EPSG:32630",
            x_size,
            y_size,
            area_extent,
        )
        scene1["1"] = xr.DataArray(np.zeros((y_size, x_size)), dims=("y", "x"),
                                   attrs={"area": area_def})
        # by x/y bbox
        new_scn1 = scene1.crop(xy_bbox=(719695.7781587119, 5427887.407618969, 725068.1609052602, 5433708.364368956))
        assert "1" in new_scn1
        assert new_scn1["1"].shape == (198, 182)

    def test_crop_rgb(self):
        """Test the crop method on multi-dimensional data."""
        from pyresample.geometry import AreaDefinition
        scene1 = Scene()
        area_extent = (-5570248.477339745, -5561247.267842293, 5567248.074173927,
                       5570248.477339745)
        proj_dict = {"a": 6378169.0, "b": 6356583.8, "h": 35785831.0,
                     "lon_0": 0.0, "proj": "geos", "units": "m"}
        x_size = 3712
        y_size = 3712
        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            proj_dict,
            x_size,
            y_size,
            area_extent,
        )
        area_def2 = AreaDefinition(
            "test2",
            "test2",
            "test2",
            proj_dict,
            x_size // 2,
            y_size // 2,
            area_extent,
            )
        scene1["1"] = xr.DataArray(np.zeros((3, y_size, x_size)),
                                   dims=("bands", "y", "x"),
                                   attrs={"area": area_def})
        scene1["2"] = xr.DataArray(np.zeros((y_size // 2, 3, x_size // 2)),
                                   dims=("y", "bands", "x"),
                                   attrs={"area": area_def2})

        # by lon/lat bbox
        new_scn1 = scene1.crop(ll_bbox=(-20., -5., 0, 0))
        assert "1" in new_scn1
        assert "2" in new_scn1
        assert "bands" in new_scn1["1"].dims
        assert "bands" in new_scn1["2"].dims
        assert new_scn1["1"].shape == (3, 184, 714)
        assert new_scn1["2"].shape == (92, 3, 357)


@pytest.mark.usefixtures("include_test_etc")
class TestSceneResampling:
    """Test resampling a Scene to another Scene object."""

    def _fake_resample_dataset(self, dataset, dest_area, **kwargs):
        """Return copy of dataset pretending it was resampled."""
        return dataset.copy()

    def _fake_resample_dataset_force_20x20(self, dataset, dest_area, **kwargs):
        """Return copy of dataset pretending it was resampled to (20, 20) shape."""
        data = np.zeros((20, 20))
        attrs = dataset.attrs.copy()
        attrs["area"] = dest_area
        return xr.DataArray(
            data,
            dims=("y", "x"),
            attrs=attrs,
        )

    @mock.patch("satpy.resample.base.resample_dataset")
    @pytest.mark.parametrize("datasets", [
        None,
        ("comp13", "ds5", "ds2"),
    ])
    def test_resample_scene_copy(self, rs, datasets):
        """Test that the Scene is properly copied during resampling.

        The Scene that is created as a copy of the original Scene should not
        be able to affect the original Scene object.

        """
        from pyresample.geometry import AreaDefinition
        rs.side_effect = self._fake_resample_dataset_force_20x20

        proj_str = ("+proj=lcc +datum=WGS84 +ellps=WGS84 "
                    "+lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs")
        area_def = AreaDefinition("test", "test", "test", proj_str, 5, 5, (-1000., -1500., 1000., 1500.))
        area_def.get_area_slices = mock.MagicMock()
        scene = Scene(filenames=["fake1_1.txt", "fake1_highres_1.txt"], reader="fake1")

        scene.load(["comp19"])
        new_scene = scene.resample(area_def, datasets=datasets)
        new_scene["new_ds"] = new_scene["comp19"].copy()

        scene.load(["ds1"])

        comp19_node = scene._dependency_tree["comp19"]
        ds5_mod_id = make_dataid(name="ds5", modifiers=("res_change",))
        ds5_node = scene._dependency_tree[ds5_mod_id]
        comp13_node = scene._dependency_tree["comp13"]

        assert comp13_node.data[1][0] is comp19_node.data[1][0]
        assert comp13_node.data[1][0] is ds5_node
        pytest.raises(KeyError, scene._dependency_tree.__getitem__, "new_ds")

        # comp19 required resampling to produce so we should have its 3 deps
        # 1. comp13
        # 2. ds5
        # 3. ds2
        # Then we loaded ds1 separately so we should have
        # 4. ds1
        loaded_ids = list(scene.keys())
        assert len(loaded_ids) == 4
        for name in ("comp13", "ds5", "ds2", "ds1"):
            assert any(x["name"] == name for x in loaded_ids)

        loaded_ids = list(new_scene.keys())
        assert len(loaded_ids) == 2
        assert loaded_ids[0] == make_cid(name="comp19")
        assert loaded_ids[1] == make_cid(name="new_ds")

    @mock.patch("satpy.resample.base.resample_dataset")
    def test_resample_scene_preserves_requested_dependencies(self, rs):
        """Test that the Scene is properly copied during resampling.

        The Scene that is created as a copy of the original Scene should not
        be able to affect the original Scene object.

        """
        from pyresample.geometry import AreaDefinition

        rs.side_effect = self._fake_resample_dataset
        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            "+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs",
            5,
            5,
            (-1000., -1500., 1000., 1500.),
        )
        area_def.get_area_slices = mock.MagicMock()
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")

        # Set PYTHONHASHSEED to 0 in the interpreter to test as intended (comp26 comes before comp14)
        scene.load(["comp26", "comp14"], generate=False)
        scene.resample(area_def, unload=True)
        new_scene_2 = scene.resample(area_def, unload=True)

        assert "comp14" not in scene
        assert "comp26" not in scene
        assert "comp14" in new_scene_2
        assert "comp26" in new_scene_2
        assert "ds1" not in new_scene_2  # unloaded

    @mock.patch("satpy.resample.base.resample_dataset")
    def test_resample_reduce_data_toggle(self, rs):
        """Test that the Scene can be reduced or not reduced during resampling."""
        from pyresample.geometry import AreaDefinition

        rs.side_effect = self._fake_resample_dataset_force_20x20
        proj_str = ("+proj=lcc +datum=WGS84 +ellps=WGS84 "
                    "+lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs")
        target_area = AreaDefinition("test", "test", "test", proj_str, 4, 4, (-1000., -1500., 1000., 1500.))
        area_def = AreaDefinition("test", "test", "test", proj_str, 5, 5, (-1000., -1500., 1000., 1500.))
        area_def.get_area_slices = mock.MagicMock()
        get_area_slices = area_def.get_area_slices
        get_area_slices.return_value = (slice(0, 3, None), slice(0, 3, None))
        area_def_big = AreaDefinition("test", "test", "test", proj_str, 10, 10, (-1000., -1500., 1000., 1500.))
        area_def_big.get_area_slices = mock.MagicMock()
        get_area_slices_big = area_def_big.get_area_slices
        get_area_slices_big.return_value = (slice(0, 6, None), slice(0, 6, None))

        # Test that data reduction can be disabled
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["comp19"])
        scene["comp19"].attrs["area"] = area_def
        scene["comp19_big"] = xr.DataArray(
            da.zeros((10, 10)), dims=("y", "x"),
            attrs=scene["comp19"].attrs.copy())
        scene["comp19_big"].attrs["area"] = area_def_big
        scene["comp19_copy"] = scene["comp19"].copy()
        orig_slice_data = scene._slice_data
        # we force the below order of processing to test that success isn't
        # based on data of the same resolution being processed together
        test_order = [
            make_cid(**scene["comp19"].attrs),
            make_cid(**scene["comp19_big"].attrs),
            make_cid(**scene["comp19_copy"].attrs),
        ]
        with mock.patch("satpy.scene.Scene._slice_data") as slice_data, \
                mock.patch("satpy.dataset.dataset_walker") as ds_walker:
            ds_walker.return_value = test_order
            slice_data.side_effect = orig_slice_data
            scene.resample(target_area, reduce_data=False)
            slice_data.assert_not_called()
            get_area_slices.assert_not_called()
            scene.resample(target_area)
            assert slice_data.call_count == 3
            assert get_area_slices.call_count == 1
            assert get_area_slices_big.call_count == 1
            scene.resample(target_area, reduce_data=True)
            # 2 times for each dataset
            # once for default (reduce_data=True)
            # once for kwarg forced to `True`
            assert slice_data.call_count == 2 * 3
            # get area slices is called again, once per area
            assert get_area_slices.call_count == 2
            assert get_area_slices_big.call_count == 2

    def test_resample_ancillary(self):
        """Test that the Scene reducing data does not affect final output."""
        from pyresample.geometry import AreaDefinition

        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            "+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs",
            5,
            5,
            (-1000., -1500., 1000., 1500.),
        )
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")

        scene.load(["comp19", "comp20"])
        scene["comp19"].attrs["area"] = area_def
        scene["comp19"].attrs["ancillary_variables"] = [scene["comp20"]]
        scene["comp20"].attrs["area"] = area_def

        dst_area = AreaDefinition(
            "dst",
            "dst",
            "dst",
            "+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs",
            2,
            2,
            (-1000., -1500., 0., 0.),
        )
        new_scene = scene.resample(dst_area)
        assert new_scene["comp20"] is new_scene["comp19"].attrs["ancillary_variables"][0]

    def test_resample_multi_ancillary(self):
        """Test that multiple ancillary variables are retained after resampling.

        This test corresponds to GH#2329
        """
        from pyresample import create_area_def
        sc = Scene()
        n = 5
        ar = create_area_def("a", 4087, resolution=1000, center=(0, 0), shape=(n, n))
        anc_vars = [xr.DataArray(
            np.arange(n*n).reshape(n, n)*i,
            dims=("y", "x"),
            attrs={"name": f"anc{i:d}", "area": ar}) for i in range(2)]
        sc["test"] = xr.DataArray(
            np.arange(n*n).reshape(n, n),
            dims=("y", "x"),
            attrs={
                "area": ar,
                "name": "test",
                "ancillary_variables": anc_vars})
        subset = create_area_def("b", 4087, resolution=800, center=(0, 0),
                                 shape=(n-1, n-1))
        ls = sc.resample(subset)
        assert ([av.attrs["name"] for av in sc["test"].attrs["ancillary_variables"]] ==
                [av.attrs["name"] for av in ls["test"].attrs["ancillary_variables"]])

    def test_resample_reduce_data(self):
        """Test that the Scene reducing data does not affect final output."""
        from pyresample.geometry import AreaDefinition
        proj_str = ("+proj=lcc +datum=WGS84 +ellps=WGS84 "
                    "+lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs")
        area_def = AreaDefinition("test", "test", "test", proj_str, 20, 20, (-1000., -1500., 1000., 1500.))
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")

        scene.load(["comp19"])
        scene["comp19"].attrs["area"] = area_def
        dst_area = AreaDefinition("dst", "dst", "dst",
                                  proj_str,
                                  20, 20,
                                  (-1000., -1500., 0., 0.),
                                  )
        new_scene1 = scene.resample(dst_area, reduce_data=False)
        new_scene2 = scene.resample(dst_area)
        new_scene3 = scene.resample(dst_area, reduce_data=True)
        assert new_scene1["comp19"].shape == (20, 20, 3)
        assert new_scene2["comp19"].shape == (20, 20, 3)
        assert new_scene3["comp19"].shape == (20, 20, 3)

    @mock.patch("satpy.resample.base.resample_dataset")
    def test_no_generate_comp10(self, rs):
        """Test generating a composite after loading."""
        from pyresample.geometry import AreaDefinition

        rs.side_effect = self._fake_resample_dataset
        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            "+proj=lcc +datum=WGS84 +ellps=WGS84 +lon_0=-95. +lat_0=25 +lat_1=25 +units=m +no_defs",
            200,
            400,
            (-1000., -1500., 1000., 1500.),
        )

        # it is fine that an optional prereq doesn't exist
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["comp10"], generate=False)
        assert any(ds_id["name"] == "comp10" for ds_id in scene._wishlist)
        assert "comp10" not in scene
        # two dependencies should have been loaded
        assert len(scene._datasets) == 2
        assert len(scene.missing_datasets) == 1

        new_scn = scene.resample(area_def, generate=False)
        assert "comp10" not in scene
        # two dependencies should have been loaded
        assert len(scene._datasets) == 2
        assert len(scene.missing_datasets) == 1

        new_scn._generate_composites_from_loaded_datasets()
        assert any(ds_id["name"] == "comp10" for ds_id in new_scn._wishlist)
        assert "comp10" in new_scn
        assert not new_scn.missing_datasets

        # try generating them right away
        new_scn = scene.resample(area_def)
        assert any(ds_id["name"] == "comp10" for ds_id in new_scn._wishlist)
        assert "comp10" in new_scn
        assert not new_scn.missing_datasets

    def test_comp_loading_after_resampling_existing_sensor(self):
        """Test requesting a composite after resampling."""
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene.load(["ds1", "ds2"])
        new_scn = scene.resample(resampler="native")

        # Can't load from readers after resampling
        with pytest.raises(KeyError):
            new_scn.load(["ds3"])

        # But we can load composites because the sensor composites were loaded
        # when the reader datasets were accessed
        new_scn.load(["comp2"])
        assert "comp2" in new_scn

    def test_comp_loading_after_resampling_new_sensor(self):
        """Test requesting a composite after resampling when the sensor composites weren't loaded before."""
        # this is our base Scene with sensor "fake_sensor2"
        scene1 = Scene(filenames=["fake2_3ds_1.txt"], reader="fake2_3ds")
        scene1.load(["ds2"])
        new_scn = scene1.resample(resampler="native")

        # Can't load from readers after resampling
        with pytest.raises(KeyError):
            new_scn.load(["ds3"])

        # Can't load the composite from fake_sensor composites yet
        # 'ds1' is missing
        with pytest.raises(KeyError):
            new_scn.load(["comp2"])

        # artificial DataArray "created by the user"
        # mimics a user adding their own data with the same sensor
        user_da = scene1["ds2"].copy()
        user_da.attrs["name"] = "ds1"
        user_da.attrs["sensor"] = {"fake_sensor2"}
        # Add 'ds1' that doesn't provide the 'fake_sensor' sensor
        new_scn["ds1"] = user_da
        with pytest.raises(KeyError):
            new_scn.load(["comp2"])
        assert "comp2" not in new_scn

        # artificial DataArray "created by the user"
        # mimics a user adding their own data with its own sensor to the Scene
        user_da = scene1["ds2"].copy()
        user_da.attrs["name"] = "ds1"
        user_da.attrs["sensor"] = {"fake_sensor"}
        # Now 'fake_sensor' composites have been loaded
        new_scn["ds1"] = user_da
        new_scn.load(["comp2"])
        assert "comp2" in new_scn

    def test_comp_loading_multisensor_composite_created_user(self):
        """Test that multisensor composite can be created manually.

        Test that if the user has created datasets "manually", that
        multi-sensor composites provided can still be read.
        """
        scene1 = Scene(filenames=["fake1_1.txt"], reader="fake1")
        scene1.load(["ds1"])
        scene2 = Scene(filenames=["fake4_1.txt"], reader="fake4")
        scene2.load(["ds4_b"])
        scene3 = Scene()
        scene3["ds1"] = scene1["ds1"]
        scene3["ds4_b"] = scene2["ds4_b"]
        scene3.load(["comp_multi"])
        assert "comp_multi" in scene3

    def test_comps_need_resampling_optional_mod_deps(self):
        """Test that a composite with complex dependencies.

        This is specifically testing the case where a compositor depends on
        multiple resolution prerequisites which themselves are composites.
        These sub-composites depend on data with a modifier that only has
        optional dependencies. This is a very specific use case and is the
        simplest way to present the problem (so far).

        The general issue is that the Scene loading creates the "ds13"
        dataset which already has one modifier on it. The "comp27"
        composite requires resampling so its 4 prerequisites + the
        requested "ds13" (from the reader which includes mod1 modifier)
        remain. If the DependencyTree is not copied properly in this
        situation then the new Scene object will have the composite
        dependencies without resolution in its dep tree, but have
        the DataIDs with the resolution in the dataset dictionary.
        This all results in the Scene trying to regenerate composite
        dependencies that aren't needed which fail.

        """
        scene = Scene(filenames=["fake1_1.txt"], reader="fake1")
        # should require resampling
        scene.load(["comp27", "ds13"])
        assert "comp27" not in scene
        assert "ds13" in scene

        new_scene = scene.resample(resampler="native")
        assert len(list(new_scene.keys())) == 2
        assert "comp27" in new_scene
        assert "ds13" in new_scene


class TestSceneAggregation:
    """Test the scene's aggregate method."""

    def test_aggregate(self):
        """Test the aggregate method."""
        x_size = 3712
        y_size = 3712

        scene1 = self._create_test_data(x_size, y_size)

        scene2 = scene1.aggregate(func="sum", x=2, y=2)
        expected_aggregated_shape = (y_size / 2, x_size / 2)
        self._check_aggregation_results(expected_aggregated_shape, scene1, scene2, x_size, y_size)

    def test_custom_aggregate(self):
        """Test the aggregate method with custom function."""
        x_size = 3712
        y_size = 3712

        scene1 = self._create_test_data(x_size, y_size)

        scene2 = scene1.aggregate(func=np.sum, x=2, y=2)
        expected_aggregated_shape = (y_size / 2, x_size / 2)
        self._check_aggregation_results(expected_aggregated_shape, scene1, scene2, x_size, y_size)

    @staticmethod
    def _create_test_data(x_size, y_size):
        from pyresample.geometry import AreaDefinition
        scene1 = Scene()
        area_extent = (-5570248.477339745, -5561247.267842293, 5567248.074173927,
                       5570248.477339745)
        proj_dict = {"a": 6378169.0, "b": 6356583.8, "h": 35785831.0,
                     "lon_0": 0.0, "proj": "geos", "units": "m"}
        area_def = AreaDefinition(
            "test",
            "test",
            "test",
            proj_dict,
            x_size,
            y_size,
            area_extent,
        )
        scene1["1"] = xr.DataArray(np.ones((y_size, x_size)),
                                   attrs={"_satpy_id_keys": default_id_keys_config})
        scene1["2"] = xr.DataArray(np.ones((y_size, x_size)),
                                   dims=("y", "x"),
                                   attrs={"_satpy_id_keys": default_id_keys_config})
        scene1["3"] = xr.DataArray(np.ones((y_size, x_size)),
                                   dims=("y", "x"),
                                   attrs={"area": area_def, "_satpy_id_keys": default_id_keys_config})
        scene1["4"] = xr.DataArray(np.ones((y_size, x_size)),
                                   dims=("y", "x"),
                                   attrs={"area": area_def, "standard_name": "backscatter",
                                          "_satpy_id_keys": default_id_keys_config})
        return scene1

    def _check_aggregation_results(self, expected_aggregated_shape, scene1, scene2, x_size, y_size):
        assert scene1["1"] is scene2["1"]
        assert scene1["2"] is scene2["2"]
        np.testing.assert_allclose(scene2["3"].data, 4)
        assert scene2["1"].shape == (y_size, x_size)
        assert scene2["2"].shape == (y_size, x_size)
        assert scene2["3"].shape == expected_aggregated_shape
        assert "standard_name" in scene2["4"].attrs
        assert scene2["4"].attrs["standard_name"] == "backscatter"

    def test_aggregate_with_boundary(self):
        """Test aggregation with boundary argument."""
        x_size = 3711
        y_size = 3711

        scene1 = self._create_test_data(x_size, y_size)

        with pytest.raises(ValueError, match="Could not coarsen a dimension.*"):
            scene1.aggregate(func="sum", x=2, y=2, boundary="exact")

        scene2 = scene1.aggregate(func="sum", x=2, y=2, boundary="trim")
        expected_aggregated_shape = (y_size // 2, x_size // 2)
        self._check_aggregation_results(expected_aggregated_shape, scene1, scene2, x_size, y_size)
